# flake8: noqa
# pylint: disable=C0301

# Some common builtin exeptions we might want to watch for
PYTHON_BUILTIN_EXCEPTIONS = [
    "ArithmeticError",
    "AssertionError",
    "AttributeError",
#    "BaseException",
#    "BlockingIOError",
#    "BrokenPipeError",
#    "BufferError",
    "BytesWarning",
#    "ChildProcessError",
#    "ConnectionAbortedError",
#    "ConnectionError",
#    "ConnectionRefusedError",
    "ConnectionResetError",
#    "DeprecationWarning",
#    "EnvironmentError",
    "EOFError",
#    "Exception",
    "FileExistsError",
    "FileNotFoundError",
    "FloatingPointError",
#    "FutureWarning",
#    "GeneratorExit",
    "ImportError",
    "ImportWarning",
#    "IndentationError",
    "IndexError",
    "InterruptedError",
    "IOError",
    "IsADirectoryError",
#    "KeyboardInterrupt",
    "KeyError",
    "LookupError",
    "MemoryError",
    "ModuleNotFoundError",
    "NameError",
#    "NotADirectoryError",
    "NotImplementedError",
    "OSError",
    "OverflowError",
#    "PendingDeprecationWarning",
    "PermissionError",
    "ProcessLookupError",
    "RecursionError",
    "ReferenceError",
    "ResourceWarning",
    "RuntimeError",
#    "RuntimeWarning",
#    "StopAsyncIteration",
#    "StopIteration",
    "SyntaxError",
#    "SyntaxWarning",
    "SystemError",
    "SystemExit",
#    "TabError",
    "TimeoutError",
    "TypeError",
    "UnboundLocalError",
    "UnicodeDecodeError",
    "UnicodeEncodeError",
    "UnicodeError",
    "UnicodeTranslateError",
    "UnicodeWarning",
    "UserWarning",
    "ValueError",
#    "Warning",
#    "WindowsError",
    "ZeroDivisionError",
]

# https://opendev.org/openstack/oslo.db/src/branch/master/oslo_db/exception.py
OSLO_DB_EXCEPTIONS = [
    "DBError",
    "DBDuplicateEntry",
    "DBConstraintError",
    "DBReferenceError",
    "DBNonExistentConstraint",
    "DBNonExistentTable",
    "DBNonExistentDatabase",
    "DBDeadlock",
    "DBInvalidUnicodeParameter",
    "DBMigrationError",
    "DBConnectionError",
    "DBDataError",
    "DBNotSupportedError",
    "InvalidSortKey",
    "ColumnError",
    "BackendNotAvailable",
    "RetryRequest",
    "NoEngineContextEstablished",
    "ContextNotRequestedError",
    "CantStartEngineError",
]

# https://opendev.org/openstack/oslo.messaging/src/branch/master/oslo_messaging/exceptions.py
OSLO_MESSAGING_EXCEPTIONS = [
    "MessagingException",
    "MessagingTimeout",
    "MessageDeliveryFailure",
    "InvalidTarget",
    "MessageUndeliverable",
]

# Exceptions common to any project should be defined here
EXCEPTIONS_COMMON = [
    r'AMQP server on .+ is unreachable',
    r'amqp.exceptions.ConnectionForced',
    r'OSError: Server unexpectedly closed connection',
] + OSLO_DB_EXCEPTIONS \
  + OSLO_MESSAGING_EXCEPTIONS \
  + PYTHON_BUILTIN_EXCEPTIONS



#!/usr/bin/python3
import html
import json
import logging
import os
import shutil
import tempfile
from typing import Literal

# load all plugins
import hotsos.plugin_extensions  # noqa: F401, pylint: disable=W0611
from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers.cli import CLIHelper
from hotsos.core.issues import IssuesManager
from hotsos.core.log import log
from hotsos.core import plugintools
from hotsos.core.exceptions import UnsupportedFormatError


class HotSOSSummary(plugintools.PluginPartBase):
    """
    This plugin will always be run and provides information specific to hotsos
    itself.
    """
    plugin_name = 'hotsos'
    plugin_root_index = 0
    summary_part_index = 0

    @classmethod
    def is_runnable(cls):
        return True

    @property
    def summary(self):
        out = {'version': HotSOSConfig.hotsos_version,
               'repo-info': HotSOSConfig.repo_info}
        if HotSOSConfig.force_mode:
            out['force'] = True

        return out


FILTER_SCHEMA = [IssuesManager.SUMMARY_OUT_ISSUES_ROOT,
                 IssuesManager.SUMMARY_OUT_BUGS_ROOT]


SUPPORTED_SUMMARY_FORMATS = ['yaml', 'json', 'markdown', 'html']
SUPPORTED_MINIMAL_MODES = ['full', 'short', 'very-short']


class OutputBuilder:
    """Builder class for generating desired output format from
    raw dictionary."""

    def __init__(self, content):
        self.content = content

    @staticmethod
    def _minimise(summary, mode):
        """ Converts the master output to include only issues and bugs. """

        log.debug("Minimising output (mode=%s).", mode)

        if not summary:
            return summary

        if mode == 'short':
            return OutputBuilder._get_short_format(summary)
        if mode == 'very-short':
            return OutputBuilder._get_very_short_format(summary)

        log.warning("Unknown minimalmode '%s'", mode)
        return summary

    @staticmethod
    def _get_short_format(summary):
        filtered = {}
        for plugin in summary:
            for key in FILTER_SCHEMA:
                if key not in summary[plugin]:
                    continue

                if key not in filtered:
                    filtered[key] = {}

                items = summary[plugin][key]
                filtered[key][plugin] = items

        return filtered

    @staticmethod
    def _get_very_short_format(summary):
        filtered = {}
        for plugin in summary:
            for key in FILTER_SCHEMA:
                if key not in summary[plugin]:
                    continue

                if key not in filtered:
                    filtered[key] = {}

                items = summary[plugin][key]
                if isinstance(items, dict):
                    filtered[key][plugin] = {key: len(val)
                                             for key, val in items.items()}
                    continue

                # support old format summaries
                if key == IssuesManager.SUMMARY_OUT_ISSUES_ROOT:
                    aggr_info = {}
                else:
                    aggr_info = []

                for item in items:
                    if key == IssuesManager.SUMMARY_OUT_ISSUES_ROOT:
                        item_key = item['type']
                        if item_key not in aggr_info:
                            aggr_info[item_key] = 1
                        else:
                            aggr_info[item_key] += 1

                    else:
                        item_key = item['id']
                        aggr_info.append(item_key)

                filtered[key][plugin] = aggr_info

        return filtered

    def filter(self, plugin_name=None):
        if plugin_name:
            self.content = {plugin_name: self.content[plugin_name]}
        return self

    def minimal(self, mode=None):
        if mode:
            self.content = self._minimise(self.content, mode)
        return self

    def to(self, fmt: Literal[SUPPORTED_SUMMARY_FORMATS],
           **kwargs):
        if fmt == "html":
            return self.to_html(**kwargs)
        if fmt == "json":
            return self.to_json()
        if fmt == "yaml":
            return self.to_yaml()
        if fmt == "markdown":
            return self.to_markdown()

        raise UnsupportedFormatError(fmt)

    def to_html(self, *, max_level=2, html_escape=False):
        hostname = CLIHelper().hostname() or ""
        result = plugintools.HTMLFormatter(
            hostname=hostname,
            max_level=max_level
        ).dump(self.content)

        return result if not html_escape else html.escape(result)

    def to_json(self):
        return json.dumps(self.content, indent=2, sort_keys=True)

    def to_yaml(self):
        return plugintools.yaml_dump(self.content)

    def to_markdown(self):
        return plugintools.MarkdownFormatter().dump(self.content)


class OutputManager():
    """ Handle conversion of plugin output into summary format. """

    def __init__(self, initial=None):
        self._summary = initial or {}

    def get_builder(self):
        return OutputBuilder(self._summary)

    @staticmethod
    def _save_to_file(path, content):
        with open(path, "w", encoding="utf-8") as fd:
            fd.write(content)
            fd.write("\n")

    def save(self, name, html_escape=False, output_path=None):
        """
        Save all formats and styles to disk using either the provided path or
        an autogenerated one.

        Returns path of saved data.
        """
        if output_path:
            output_root = output_path
        else:
            output_root = f"hotsos-output-{CLIHelper().date(format='+%s')}"

        for minimal_mode in SUPPORTED_MINIMAL_MODES:
            _minimal_mode = minimal_mode.replace('-', '_')
            for fmt in SUPPORTED_SUMMARY_FORMATS:
                output_path = os.path.join(output_root, name, 'summary',
                                           _minimal_mode, fmt)
                if minimal_mode == 'full':
                    minimal_mode = None

                if not os.path.exists(output_path):
                    os.makedirs(output_path)

                # Save per-plugin summary
                for plugin in self._summary:
                    path = os.path.join(output_path,
                                        f"hotsos-summary.{plugin}.{fmt}")
                    output = self.get_builder()
                    output.filter(plugin).minimal(minimal_mode)
                    formatted_output = output.to(
                        fmt=fmt,
                        html_escape=html_escape)
                    log.debug('Saving plugin %s summary as %s',
                              plugin,
                              fmt)
                    self._save_to_file(path, formatted_output)

                # Save all summary
                path = os.path.join(output_path, f"hotsos-summary.all.{fmt}")
                output = self.get_builder()
                output.minimal(minimal_mode)
                formatted_output = output.to(fmt=fmt, html_escape=html_escape)
                log.debug('Saving all summary as %s', fmt)
                self._save_to_file(path, formatted_output)

                if not minimal_mode:
                    dst = os.path.join(output_root, f'{name}.summary.{fmt}')
                    if os.path.exists(dst):
                        os.remove(dst)

                    os.symlink(path.partition(output_root)[2].lstrip('/'), dst)

        if log.handlers and isinstance(log.handlers[0], logging.FileHandler):
            log.handlers[0].close()
            # no logging after this point
            shutil.move(log.handlers[0].baseFilename,
                        os.path.join(output_root, name, 'hotsos.log'))

        return output_root

    def update(self, plugin, content):
        self._summary[plugin] = content


class HotSOSClient():
    """
    Main HotSOS client from which all plugins are run.
    """
    def __init__(self, plugins=None):
        """
        @param plugins: list of plugin names to run. If no plugins are provided
        all will be run.
        """
        self._summary = OutputManager()
        self.plugins = plugins or plugintools.PLUGINS.keys()

    @staticmethod
    def setup_global_env():
        """ State saved here persists across all plugin runs. """
        log.debug("setting up global env")
        global_tmp_dir = tempfile.mkdtemp()
        HotSOSConfig.global_tmp_dir = global_tmp_dir
        os.makedirs(os.path.join(global_tmp_dir, 'locks'))

    @staticmethod
    def teardown_global_env():
        log.debug("tearing down global env")
        if os.path.exists(HotSOSConfig.global_tmp_dir):
            shutil.rmtree(HotSOSConfig.global_tmp_dir)
        # Ensure tmp dir doesn't get accidentally recreated
        HotSOSConfig.plugin_tmp_dir = None

    @staticmethod
    def setup_plugin_env(plugin):
        """ State saved here is specific to a plugin. """
        log.debug("setting up plugin env")
        global_tmp = HotSOSConfig.global_tmp_dir
        HotSOSConfig.plugin_tmp_dir = tempfile.mkdtemp(prefix=plugin,
                                                       dir=global_tmp)

    @property
    def summary(self):
        return self._summary

    def run(self):
        """
        Run the selected plugins. This will run the automatic (defs) checks as
        well as any extensions.
        """
        log.name = 'hotsos.client'
        try:
            self.setup_global_env()
            for plugin in plugintools.get_plugins_sorted():
                if plugin not in self.plugins:
                    continue

                self.setup_plugin_env(plugin)
                log.name = f'hotsos.plugin.{plugin}'
                log.debug("running plugin %s", plugin)
                HotSOSConfig.plugin_name = plugin
                content = plugintools.PluginRunner(plugin).run()
                if content:
                    self.summary.update(plugin, content.get(plugin))
        finally:
            log.name = 'hotsos.client'
            self.teardown_global_env()



import os
import re
from functools import cached_property

from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers.config import GenericIniConfig
from hotsos.core.host_helpers import (
    CLIHelper, CLIHelperFile,
    SYSCtlFactory, UptimeHelper
)
from hotsos.core.log import log
from hotsos.core.search import (
    FileSearcher, SearchDef,
    SequenceSearchDef
)


class NUMAInfo():
    """ Interface to host NUMA information. """
    numactl = ""

    def __init__(self):
        try:
            self.numactl = CLIHelper().numactl() or ""
        except OSError:
            self.numactl = ""

        self._nodes = {}

    @cached_property
    def nodes(self):
        """Returns dictionary of numa nodes and their associated list of cpu
           cores.
        """
        if self._nodes:
            return self._nodes

        node_ids = []
        for line in self.numactl:
            expr = r'^available:\s+[0-9]+\s+nodes\s+\(([0-9\-]+)\)'
            ret = re.compile(expr).match(line)
            if ret:
                p = ret[1].partition('-')
                if p[1] == '-':
                    node_ids = range(int(p[0]), int(p[2]) + 1)
                else:
                    node_ids = [int(p[0])]

                break

        for node in node_ids:
            for line in self.numactl:
                expr = rf'^node\s+{node}\s+cpus:\s([0-9\s]+)'
                ret = re.compile(expr).match(line)
                if ret:
                    self._nodes[node] = [int(e) for e in ret[1].split()]
                    break

        return self._nodes

    def cores(self, node=None):
        """Returns list of cores for a given numa node.

        If no node id is provided, all cores from all numa nodes are returned.
        """
        if not self.nodes:
            return []

        if node is None:
            _cores = []
            for c in self.nodes.values():
                _cores += c

            return _cores

        return self.nodes.get(node)


class SystemBase():
    """ Base class for system checks. """
    @cached_property
    def date(self):
        return CLIHelper().date(no_format=True)

    @cached_property
    def hostname(self):
        return CLIHelper().hostname()

    @cached_property
    def uptime(self):
        return str(UptimeHelper())

    @cached_property
    def os_release_name(self):
        data_source = os.path.join(HotSOSConfig.data_root, "etc/lsb-release")
        if not os.path.exists(data_source):
            return None

        with open(data_source, encoding='utf-8') as fd:
            for line in fd.read().split():
                ret = re.compile(r"^DISTRIB_CODENAME=(.+)").match(line)
                if ret:
                    return f"ubuntu {ret[1]}"

        return None

    @cached_property
    def virtualisation_type(self):
        """
        @return: virt type e.g. kvm or lxc if host is virtualised otherwise
                 None.
        """
        info = CLIHelper().hostnamectl()
        for line in info:
            split_line = line.partition(': ')
            if 'Virtualization' in split_line[0]:
                return split_line[2].strip()

        return None

    @cached_property
    def num_cpus(self):
        """ Return number of cpus or 0 if none found. """
        lscpu_output = CLIHelper().lscpu()
        if lscpu_output:
            for line in lscpu_output:
                ret = re.compile(r"^CPU\(s\):\s+([0-9]+)\s*.*").match(line)
                if ret:
                    return int(ret[1])

        return 0

    @cached_property
    def unattended_upgrades_enabled(self):
        apt_config_dump = CLIHelper().apt_config_dump()
        if not apt_config_dump:
            return None

        for line in apt_config_dump:
            ret = re.compile(r"^APT::Periodic::Unattended-Upgrade\s+"
                             "\"([0-9]+)\";").match(line)
            if ret:
                if int(ret[1]) == 0:
                    return False

                return True

        return False

    @cached_property
    def ubuntu_pro_status(self):
        """Parse and retrieve Ubuntu Pro status

        Returns:
            Dictionary: Ubuntu pro status as a dictionary, e.g.::
            {
                "status": "<attached|not-attached|error>"
                "services": {
                    "esm-apps": {
                        "entitled": "yes",
                        "status": "enabled"
                    }, /* ... */
                },
                "account": "ACME Corporation",
                "subscription": "Ubuntu Pro (Apps-only) - Virtual",
                "technical_support_level": "essential",
                "valid_until": "Sat Jan  1 01:01:01 9999 +03"
            }

        """

        # TODO(mkg): Unfortunately, sos does not capture `pro status
        # --format json` output at the moment. We have to parse the
        # human-readable output for now. This function should ideally
        # rely on the json output when the upstream sos starts to
        # include it.
        s = FileSearcher()
        service_status_seqdef = SequenceSearchDef(
            start=SearchDef(r"^SERVICE +ENTITLED +STATUS +DESCRIPTION"),
            body=SearchDef(r"^(\S+) +(\S+) +(\S+) +(.+)\n"),
            end=SearchDef(r"\n"),
            tag="service-status")
        account_status_seqdef = SequenceSearchDef(
            start=SearchDef(r" *?(Account): (.*)\n"),
            body=SearchDef(r" *?([\S ]+): (.*)\n"),
            end=SearchDef(r" *?(Technical support level): (.*)\n"),
            tag="account-status")
        not_attached_def = SearchDef(
            r".*not attached to.*(Ubuntu (Pro|Advantage)|UA).*",
            tag="not-attached")

        with CLIHelperFile() as cli:
            f = cli.pro_status()
            s.add(not_attached_def, f)
            s.add(service_status_seqdef, f)
            s.add(account_status_seqdef, f)
            results = s.run()
            if results.find_by_tag("not-attached"):
                return {"status": "not-attached"}

            ssects = results.find_sequence_sections(service_status_seqdef)
            asects = results.find_sequence_sections(account_status_seqdef)
            if not all([ssects, asects]):
                log.debug("badness: `pro status` does not match "
                          "the expected format")
                return {"status": "error"}

            result = {}
            result["status"] = "attached"
            result["services"] = {}

            for values in ssects.values():
                result["services"] = {**result["services"], **{
                    v.get(1): {
                        "entitled": v.get(2),
                        "status": v.get(3)
                    }
                    for v in values
                    if v.tag == service_status_seqdef.body_tag
                }}

            for values in asects.values():
                result = {**result, **{
                    re.sub(r'\W+', '_',
                           v.get(1).strip()).lower(): v.get(2).strip()
                    for v in values
                }}

            return result

    @cached_property
    def sysctl_all(self):
        return SYSCtlFactory().sysctl_all


class SSSD():
    """ SSSD interface. """
    def __init__(self):
        self.sssd_config = GenericIniConfig(
            os.path.join(HotSOSConfig.data_root,
                         "etc/sssd/sssd.conf"))

    @property
    def tokengroups_enabled_domains(self):
        # If not explicitly closed
        if not self.sssd_config:
            return []

        # Filter domain sections
        domain_sections = [
            x for x in self.sssd_config.all_sections if x.startswith("domain")
        ]

        ad_domain_sections = [
            x for x in domain_sections
            if self.sssd_config.get("id_provider", x) == "ad"]

        ad_domain_sects_w_tokengroups = [
            x for x in ad_domain_sections
            if self.sssd_config.get("ldap_use_tokengroups", x) is not False]

        return ad_domain_sects_w_tokengroups



import abc
import os

from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers import CLIHelper, HostNetworkingHelper
from hotsos.core.search import (
    CommonTimestampMatcher,
    FileSearcher,
    SearchConstraintSearchSince,
)

KERNLOG_TS = r'\[\s*\d+\.\d+\]'
KERNLOG_PREFIX = rf'(?:\S+\s+\d+\s+[\d:]+\s+\S+\s+\S+:\s+)?{KERNLOG_TS}'


class CallTraceHeuristicBase():
    """ Defines a common interface for heuristics implementations. """

    @abc.abstractmethod
    def __call__(self):
        """ Ensure callable to get results. """


class CallTraceStateBase():
    """
    A state capture object that allows getting and setting arbitrary state.
    """

    def __init__(self):
        self._info = {}

    def add(self, key, value):
        self._info[key] = value

    def __getattr__(self, key):
        return self._info.get(key)


class TraceTypeBase(abc.ABC):
    """
    This defines a common interface to trace types and should be implemented
    for any trace types we want to capture. Implementations of this object
    are typically registered with a CallTraceManager for processing.
    """

    @property
    @abc.abstractmethod
    def name(self):
        """
        Name used to identify the type of call trace e.g. "oomkiller".
        """

    @property
    @abc.abstractmethod
    def searchdef(self):
        """
        A search definition object (simple or sequence) used to identify this
        call trace.
        """

    @abc.abstractmethod
    def apply(self, results):
        """
        Take results and parse into constituent parts.

        @param results: a SearchResultsCollection object containing the results
        of our search.
        """

    @property
    @abc.abstractmethod
    def heuristics(self):
        """ Return a list of CallTraceHeuristic objects that can be used to
        run checks on any identified call traces. """

    @abc.abstractmethod
    def __len__(self):
        """ Return number of call stacks identified.  """

    @abc.abstractmethod
    def __iter__(self):
        """ Iterate over each call trace found. """


class KernLogBase():
    """ Base class for kernlog analysis implementations. """
    def __init__(self):
        c = SearchConstraintSearchSince(ts_matcher_cls=CommonTimestampMatcher)
        self.searcher = FileSearcher(constraint=c)
        self.hostnet_helper = HostNetworkingHelper()
        self.cli_helper = CLIHelper()

    @property
    def path(self):
        path = os.path.join(HotSOSConfig.data_root, 'var/log/kern.log')
        if HotSOSConfig.use_all_logs:
            return f"{path}*"

        return path



from hotsos.core.plugins.pacemaker import PacemakerChecks
from hotsos.core.plugintools import (
    summary_entry,
    get_min_available_entry_index,
)


class PacemakerSummary(PacemakerChecks):
    """ Implementation of Pacemaker summary. """
    summary_part_index = 0

    # REMINDER: common entries are implemented in the SummaryBase base class
    #           and only application plugin specific customisations are
    #           implemented here. We use the get_min_available_entry_index() to
    #           ensure that additional entries don't clobber existing ones but
    #           conversely can also replace them by re-using their indices.

    @summary_entry('nodes', get_min_available_entry_index())
    def summary_nodes(self):
        nodes = {}
        if self.online_nodes:
            nodes["online"] = self.online_nodes
        if self.offline_nodes:
            nodes["offline"] = self.offline_nodes

        return nodes or None



from hotsos.core.plugins.kubernetes import KubernetesChecks
from hotsos.core.plugintools import (
    summary_entry,
    get_min_available_entry_index,
    DefaultSummaryEntryIndexes,
)


class KubernetesSummary(KubernetesChecks):
    """ Implementation of Kubernetes summary. """
    summary_part_index = 0

    # REMINDER: common entries are implemented in the SummaryBase base class
    #           and only application plugin specific customisations are
    #           implemented here. We use the get_min_available_entry_index() to
    #           ensure that additional entries don't clobber existing ones but
    #           conversely can also replace them by re-using their indices.

    @summary_entry('dpkg', DefaultSummaryEntryIndexes.DPKG)
    def summary_dpkg(self):
        """
        Override the default entry to include all discovered packages.
        """
        return self.apt.all_formatted or None

    @summary_entry('pods', get_min_available_entry_index())
    def summary_pods(self):
        return self.pods or None

    @summary_entry('containers', get_min_available_entry_index() + 1)
    def summary_containers(self):
        return self.containers or None

    @summary_entry('flannel', get_min_available_entry_index() + 2)
    def summary_flannel(self):
        info = {}
        for port in self.flannel_ports:
            info[port.name] = port.encap_info
            if port.addresses:
                info[port.name]['addr'] = port.addresses[0]

        return info or None




# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' designate/exceptions.py
DESIGNATE_EXCEPTIONS = [
    "DesignateException",
    "Backend",
    "RelationNotLoaded",
    "AdapterNotFound",
    "NSD4SlaveBackendError",
    "NotImplemented",
    "XFRFailure",
    "ConfigurationError",
    "UnknownFailure",
    "CommunicationFailure",
    "KeystoneCommunicationFailure",
    "NeutronCommunicationFailure",
    "NoFiltersConfigured",
    "NoServersConfigured",
    "MultiplePoolsFound",
    "NoPoolTargetsConfigured",
    "OverQuota",
    "QuotaResourceUnknown",
    "InvalidObject",
    "BadAction",
    "BadRequest",
    "EmptyRequestBody",
    "InvalidProject",
    "InvalidUUID",
    "NetworkEndpointNotFound",
    "MarkerNotFound",
    "ValueError",
    "InvalidMarker",
    "InvalidSortDir",
    "InvalidLimit",
    "InvalidSortKey",
    "InvalidJson",
    "InvalidOperation",
    "UnsupportedAccept",
    "UnsupportedContentType",
    "InvalidZoneName",
    "InvalidRecordSetName",
    "InvalidRecordSetLocation",
    "InvaildZoneTransfer",
    "InvalidTTL",
    "ZoneHasSubZone",
    "Forbidden",
    "IllegalChildZone",
    "IllegalParentZone",
    "IncorrectZoneTransferKey",
    "InvalidTokenScope",
    "Duplicate",
    "DuplicateServiceStatus",
    "DuplicateQuota",
    "DuplicateServer",
    "DuplicateTsigKey",
    "DuplicateZone",
    "DuplicateTld",
    "DuplicateRecordSet",
    "DuplicateRecord",
    "DuplicateBlacklist",
    "DuplicatePool",
    "DuplicatePoolAttribute",
    "DuplicatePoolNsRecord",
    "DuplicatePoolNameserver",
    "DuplicatePoolTarget",
    "DuplicatePoolTargetOption",
    "DuplicatePoolTargetMaster",
    "DuplicatePoolAlsoNotify",
    "DuplicateZoneImport",
    "DuplicateZoneExport",
    "MethodNotAllowed",
    "DuplicateZoneTransferRequest",
    "DuplicateZoneTransferAccept",
    "DuplicateZoneAttribute",
    "DuplicateZoneMaster",
    "NotFound",
    "ServiceStatusNotFound",
    "QuotaNotFound",
    "ServerNotFound",
    "TsigKeyNotFound",
    "BlacklistNotFound",
    "ZoneNotFound",
    "ZoneMasterNotFound",
    "ZoneAttributeNotFound",
    "TldNotFound",
    "RecordSetNotFound",
    "RecordNotFound",
    "ReportNotFound",
    "PoolNotFound",
    "NoValidPoolFound",
    "PoolAttributeNotFound",
    "PoolNsRecordNotFound",
    "PoolNameserverNotFound",
    "PoolTargetNotFound",
    "PoolTargetOptionNotFound",
    "PoolTargetMasterNotFound",
    "PoolAlsoNotifyNotFound",
    "ZoneTransferRequestNotFound",
    "ZoneTransferAcceptNotFound",
    "ZoneImportNotFound",
    "ZoneExportNotFound",
    "LastServerDeleteNotAllowed",
    "ResourceNotFound",
    "MissingProjectID",
]



import os
import re
import uuid
from functools import cached_property

from hotsos.core.config import HotSOSConfig
from hotsos.core.factory import FactoryBase
from hotsos.core.log import log
from hotsos.core.host_helpers import (
    CLIHelper,
    CLIHelperFile,
    HostNetworkingHelper
)
from hotsos.core.search import (
    ExtraSearchConstraints,
    FileSearcher,
    SearchDef,
    create_constraint,
)
from hotsos.core.plugins.openvswitch.common import OpenvSwitchGlobalSearchBase


class OVSDBTable():
    """
    Provides an interface to an OVSDB table. Records can be extracted from
    either 'get' or 'list' command outputs. We try 'get' first and of not found
    we search in output of 'list'.
    """
    def __init__(self, name):
        self.name = name

    @staticmethod
    def _convert_record_to_dict(record):
        """ Convert the ovsdb record dict format to a python dictionary. """
        out = {}
        if not record or record == '{}':
            return out

        expr = r'(\S+="[^"]+"|\S+=\S+),? ?'
        for field in re.compile(expr).findall(record):
            for char in [',', '}', '{']:
                field = field.strip(char)

            key, _, val = field.partition('=')
            out[key] = val.strip('"')

        return out

    @staticmethod
    def _get_cmd(table):
        return lambda **kwargs: CLIHelper().ovs_vsctl_get(table=table,
                                                          **kwargs)

    @staticmethod
    def _list_cmd(table):
        return lambda **kwargs: CLIHelper().ovs_vsctl_list(table=table,
                                                           **kwargs)

    def _fallback_query(self, column):
        """ Find first occurrence of column and return it. """
        for cmd in [self._list_cmd(self.name),
                    self._list_cmd(self.name.lower())]:
            for line in cmd():
                if not line.startswith(f'{column} '):
                    continue

                return line.partition(':')[2].strip()

        return None

    def get(self, record, column):
        """
        Try to get column using get command and failing that, try getting it
        from list.
        """
        value = self._get_cmd(self.name)(record=record, column=column)
        if not value:
            value = self._fallback_query(column=column)

        return self._convert_record_to_dict(value)

    def __getattr__(self, column):
        """ Get column for special records i.e. with key '.' """
        return self.get(record='.', column=column)


class OVSDB(FactoryBase):
    """
    This class is used like a factory in that attributes are table names that
    return OVSDBTable objects on which attributes are row values.
    """
    def __getattr__(self, table):
        return OVSDBTable(table)


class OVSDPLookups():
    """ Interface to OVS datapath lookups. """
    def __init__(self):
        cli = CLIHelper()
        out = cli.ovs_appctl(command='dpctl/show', flags='-s',
                             args='system@ovs-system')
        cexpr = re.compile(r'\s*lookups: hit:(\S+) missed:(\S+) lost:(\S+)')
        self.fields = {'hit': 0, 'missed': 0, 'lost': 0}
        for line in out:
            ret = re.match(cexpr, line)
            if ret:
                self.fields['hit'] = int(ret.group(1))
                self.fields['missed'] = int(ret.group(2))
                self.fields['lost'] = int(ret.group(3))

    def __getattr__(self, key):
        if key in self.fields:
            return self.fields[key]

        # This is only ever expected to be used for db lookups for object attr
        # lookups so return None if key not found in DB.
        return None


class OVSBridge():
    """ Representation of an OVS bridge. """
    def __init__(self, name, nethelper):
        self.name = name
        self.cli = CLIHelper()
        self.nethelper = nethelper

    @cached_property
    def ports(self):
        ports = []
        for line in self.cli.ovs_ofctl(command='show', args=self.name):
            ret = re.compile(r'^\s+\d+\((\S+)\):\s+').match(line)
            if ret:
                name = ret.group(1)
                port = self.nethelper.get_interface_with_name(name)
                if not port:
                    port = name

                ports.append(port)

        return ports


class OpenvSwitchBase():
    """ Base class for OVS checks. """
    def __init__(self, *args, global_searcher=None, **kwargs):
        self.global_searcher = global_searcher
        super().__init__(*args, **kwargs)
        self.cli = CLIHelper()
        self.net_helper = HostNetworkingHelper()
        self.ovsdb = OVSDB()

    @cached_property
    def bridges(self):
        bridges = self.cli.ovs_vsctl_list_br()
        return [OVSBridge(br.strip(), self.net_helper) for br in bridges]

    @cached_property
    def tunnels(self):
        tunnel_info = {}
        ovn_external_ids = self.ovsdb.Open_vSwitch.external_ids
        if ovn_external_ids:
            # ovn only shows the local ip used in the db so we have to get from
            # there.
            proto = ovn_external_ids.get('ovn-encap-type')
            if proto:
                local_addr = ovn_external_ids['ovn-encap-ip']
                tunnel_info[proto] = {'local': local_addr}

        nethelp = HostNetworkingHelper()
        with CLIHelperFile() as cli:
            s = FileSearcher()
            expr = r'.+ \(([a-z]+): ([a-f\d\.:]+)->([a-f\d\.:]+), .+'
            s.add(SearchDef(expr, tag='all'),
                  cli.ovs_appctl(command='ofproto/list-tunnels'))
            results = s.run()
            for r in results.find_by_tag('all'):
                proto = r.get(1)
                if proto not in tunnel_info:
                    tunnel_info[proto] = {}

                if 'remotes' not in tunnel_info[proto]:
                    tunnel_info[proto]['remotes'] = []

                if 'local' not in tunnel_info[proto]:
                    tunnel_info[proto]['local'] = r.get(2)

                tunnel_info[proto]['remotes'].append(r.get(3))

        for proto, values in tunnel_info.items():
            local_addr = values['local']
            if local_addr:
                iface = nethelp.get_interface_with_addr(local_addr)
                if iface:
                    values['iface'] = iface.to_dict()
                    del values['local']
                else:
                    log.info("could not find interface for address '%s'",
                             local_addr)

            if 'remotes' not in values:
                log.info('no remote tunnel endpoints found for proto=%s',
                         proto)
                num_remotes = 0
            else:
                num_remotes = len(values['remotes'])

            values['remotes'] = num_remotes

        return tunnel_info

    @cached_property
    def offload_enabled(self):
        config = self.ovsdb.Open_vSwitch.other_config
        if not config:
            return False

        if config.get('hw-offload') == "true":
            return True

        return False


class OVSBFDSearch(OpenvSwitchGlobalSearchBase):
    """
    OVS BFD representation.

    This class processes the results of a search that is run using the global
    searcher.
    """
    unique_search_tag = str(uuid.uuid4()) + 'bfd'

    @classmethod
    def simple_search(cls):
        """
        Represents OVS BFD.
    
        State names: https://github.com/openvswitch/ovs/blob/ec405e8573c5ce1590171a029e6be546b3821aad/lib/bfd.c#L1058
        State transitions: https://github.com/openvswitch/ovs/blob/ec405e8573c5ce1590171a029e6be546b3821aad/lib/bfd.c#L703
        """  # noqa
        pattern = (r'([\d-]+)T([\d:]+)\.\d+Z.+\|bfd\(\S+\)\|INFO\|'
                   r'([a-z0-9-]+): BFD state change: (\S+)')
        constraint = create_constraint(search_result_age_hours=24,
                                       min_hours_since_last_boot=0)
        return SearchDef(pattern, tag=cls.unique_search_tag,
                         constraints=[constraint])

    @classmethod
    def paths(cls):
        return [os.path.join(HotSOSConfig.data_root,
                             'var/log/openvswitch/ovs-vswitchd.log')]


class OVSBFD(OpenvSwitchBase):
    """ OVS BFD representation. """
    @property
    def up_states(self):
        return ['up', 'init']

    @property
    def down_states(self):
        return ['down']

    @property
    def _results(self):
        """
        Identify a set of transitions where there are a min of 5 results within
        an hour from within the last 24 hours.
        """
        results = self.global_searcher.results.find_by_tag(
                      OVSBFDSearch.unique_search_tag)
        return ExtraSearchConstraints().apply(results, search_period_hours=1,
                                              min_results=5)

    @property
    def _transitions(self):
        """
        A transition represents going from UP to DOWN where UP can also be
        intermediary states like INIT since the remote neighbour can bring the
        local host DOWN when in this state also.
        """
        ports = {}
        for result in self._results:
            ts = f"{result.get(1)} {result.get(2)}"
            port = result.get(3)
            if port not in ports:
                ports[port] = {'changes': [(ts, result.get(4))],
                               'transitions': 0}
            else:
                ports[port]['changes'].append((ts, result.get(4)))

        for port, info in ports.items():
            for _, st in sorted(info['changes'], key=lambda e: e[0]):
                _from, _, _to = st.partition('->')
                if _to in self.down_states:
                    if _from in self.up_states:
                        info['transitions'] += 1

        return ports

    @cached_property
    def max_transitions_last_24h_within_hour(self):
        return sum((port['transitions']
                    for port in self._transitions.values()))


class OVSDPDK(OpenvSwitchBase):
    """ Interface to OVS DPDK. """
    @cached_property
    def config(self):
        return self.ovsdb.Open_vSwitch.other_config or {}

    @property
    def enabled(self):
        if self.config.get('dpdk-init') == "true":
            return True

        return False

    @property
    def pmd_cpu_mask(self):
        mask = self.config.get('pmd-cpu-mask')
        if mask is not None:
            return int(mask, 16)

        return None

    @property
    def dpdk_lcore_mask(self):
        mask = self.config.get('dpdk-lcore-mask')
        if mask is not None:
            return int(mask, 16)

        return None



# From https://github.com/libvirt/libvirt-python used by Nova
PYTHON_LIBVIRT_EXCEPTIONS = [
    "libvirtError",
]

# Since there are many many exception types we need to cherry-pick the ones we
# care about to avoid search expresssions becoming too huge.
# e.g. sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' nova/exception.py
NOVA_EXCEPTIONS = [
    "ConvertedException",
    "NovaException",
    "EncryptionFailure",
    "VirtualInterfaceCreateException",
    "VirtualInterfaceMacAddressException",
    "VirtualInterfacePlugException",
    "VirtualInterfaceUnplugException",
    "GlanceConnectionFailed",
    "CinderConnectionFailed",
    "UnsupportedCinderAPIVersion",
    "CinderAPIVersionNotAvailable",
    "Forbidden",
    "ForbiddenWithAccelerators",
    "AdminRequired",
    "PolicyNotAuthorized",
    "ImageNotActive",
    "ImageNotAuthorized",
    "Invalid",
    "InvalidConfiguration",
    "InvalidBDM",
    "InvalidBDMSnapshot",
    "InvalidBDMVolume",
    "InvalidBDMImage",
    "InvalidBDMBootSequence",
    "InvalidBDMLocalsLimit",
    "InvalidBDMEphemeralSize",
    "InvalidBDMSwapSize",
    "InvalidBDMFormat",
    "InvalidBDMForLegacy",
    "InvalidBDMVolumeNotBootable",
    "TooManyDiskDevices",
    "InvalidBDMDiskBus",
    "InvalidAttribute",
    "ValidationError",
    "VolumeAttachFailed",
    "VolumeDetachFailed",
    "MultiattachNotSupportedByVirtDriver",
    "MultiattachNotSupportedOldMicroversion",
    "MultiattachToShelvedNotSupported",
    "MultiattachSwapVolumeNotSupported",
    "VolumeNotCreated",
    "ExtendVolumeNotSupported",
    "VolumeEncryptionNotSupported",
    "VolumeTaggedAttachNotSupported",
    "VolumeTaggedAttachToShelvedNotSupported",
    "NetworkInterfaceTaggedAttachNotSupported",
    "InvalidKeypair",
    "InvalidRequest",
    "InvalidInput",
    "InvalidVolume",
    "InvalidVolumeAccessMode",
    "StaleVolumeMount",
    "InvalidMetadata",
    "InvalidMetadataSize",
    "InvalidPortRange",
    "InvalidIpProtocol",
    "InvalidContentType",
    "InvalidAPIVersionString",
    "VersionNotFoundForAPIMethod",
    "InvalidGlobalAPIVersion",
    "ApiVersionsIntersect",
    "InvalidParameterValue",
    "InvalidAggregateAction",
    "InvalidAggregateActionAdd",
    "InvalidAggregateActionDelete",
    "InvalidAggregateActionUpdate",
    "InvalidAggregateActionUpdateMeta",
    "InvalidSortKey",
    "InvalidStrTime",
    "InvalidNUMANodesNumber",
    "InvalidName",
    "InstanceInvalidState",
    "InstanceNotRunning",
    "InstanceNotInRescueMode",
    "InstanceNotRescuable",
    "InstanceNotReady",
    "InstanceSuspendFailure",
    "InstanceResumeFailure",
    "InstancePowerOnFailure",
    "InstancePowerOffFailure",
    "InstanceRebootFailure",
    "InstanceTerminationFailure",
    "InstanceDeployFailure",
    "MultiplePortsNotApplicable",
    "InvalidFixedIpAndMaxCountRequest",
    "ServiceUnavailable",
    "ServiceNotUnique",
    "ComputeResourcesUnavailable",
    "HypervisorUnavailable",
    "ComputeServiceUnavailable",
    "ComputeServiceInUse",
    "UnableToMigrateToSelf",
    "OperationNotSupportedForSEV",
    "OperationNotSupportedForVTPM",
    "OperationNotSupportedForVDPAInterface",
    "InvalidHypervisorType",
    "HypervisorTooOld",
    "DestinationHypervisorTooOld",
    "ServiceTooOld",
    "TooOldComputeService",
    "DestinationDiskExists",
    "InvalidDevicePath",
    "DevicePathInUse",
    "InvalidCPUInfo",
    "InvalidIpAddressError",
    "InvalidDiskFormat",
    "InvalidDiskInfo",
    "DiskInfoReadWriteFail",
    "ImageUnacceptable",
    "ImageBadRequest",
    "ImageImportImpossible",
    "ImageQuotaExceeded",
    "InstanceUnacceptable",
    "InvalidUUID",
    "InvalidID",
    "ConstraintNotMet",
    "NotFound",
    "VolumeAttachmentNotFound",
    "VolumeNotFound",
    "VolumeTypeNotFound",
    "UndefinedRootBDM",
    "BDMNotFound",
    "VolumeBDMNotFound",
    "VolumeBDMIsMultiAttach",
    "VolumeBDMPathNotFound",
    "DeviceDetachFailed",
    "DeviceNotFound",
    "SnapshotNotFound",
    "DiskNotFound",
    "VolumeDriverNotFound",
    "VolumeDriverNotSupported",
    "InvalidImageRef",
    "AutoDiskConfigDisabledByImage",
    "ImageNotFound",
    "ImageDeleteConflict",
    "PreserveEphemeralNotSupported",
    "InstanceMappingNotFound",
    "InvalidCidr",
    "NetworkNotFound",
    "PortNotFound",
    "NetworkNotFoundForBridge",
    "NetworkNotFoundForInstance",
    "NetworkAmbiguous",
    "UnableToAutoAllocateNetwork",
    "NetworkRequiresSubnet",
    "ExternalNetworkAttachForbidden",
    "NetworkMissingPhysicalNetwork",
    "VifDetailsMissingVhostuserSockPath",
    "VifDetailsMissingMacvtapParameters",
    "DatastoreNotFound",
    "PortInUse",
    "PortRequiresFixedIP",
    "PortNotUsable",
    "PortNotUsableDNS",
    "PortBindingFailed",
    "PortBindingDeletionFailed",
    "PortBindingActivationFailed",
    "PortUpdateFailed",
    "AttachSRIOVPortNotSupported",
    "FixedIpNotFoundForAddress",
    "FixedIpNotFoundForInstance",
    "FixedIpAlreadyInUse",
    "FixedIpAssociatedWithMultipleInstances",
    "FixedIpInvalidOnHost",
    "NoMoreFixedIps",
    "FloatingIpNotFound",
    "FloatingIpNotFoundForAddress",
    "FloatingIpMultipleFoundForAddress",
    "FloatingIpPoolNotFound",
    "NoMoreFloatingIps",
    "FloatingIpAssociated",
    "NoFloatingIpInterface",
    "FloatingIpAssociateFailed",
    "FloatingIpBadRequest",
    "KeypairNotFound",
    "ServiceNotFound",
    "ConfGroupForServiceTypeNotFound",
    "ServiceBinaryExists",
    "ServiceTopicExists",
    "HostNotFound",
    "ComputeHostNotFound",
    "HostBinaryNotFound",
    "InvalidQuotaValue",
    "InvalidQuotaMethodUsage",
    "QuotaNotFound",
    "QuotaExists",
    "QuotaResourceUnknown",
    "ProjectUserQuotaNotFound",
    "ProjectQuotaNotFound",
    "QuotaClassNotFound",
    "QuotaClassExists",
    "OverQuota",
    "SecurityGroupNotFound",
    "SecurityGroupNotFoundForProject",
    "SecurityGroupExists",
    "SecurityGroupCannotBeApplied",
    "NoUniqueMatch",
    "NoActiveMigrationForInstance",
    "MigrationNotFound",
    "MigrationNotFoundByStatus",
    "MigrationNotFoundForInstance",
    "InvalidMigrationState",
    "ConsoleLogOutputException",
    "ConsoleNotAvailable",
    "ConsoleTypeInvalid",
    "ConsoleTypeUnavailable",
    "ConsolePortRangeExhausted",
    "FlavorNotFound",
    "FlavorNotFoundByName",
    "FlavorAccessNotFound",
    "FlavorExtraSpecUpdateCreateFailed",
    "CellTimeout",
    "SchedulerHostFilterNotFound",
    "FlavorExtraSpecsNotFound",
    "ComputeHostMetricNotFound",
    "FileNotFound",
    "ClassNotFound",
    "InstanceTagNotFound",
    "KeyPairExists",
    "InstanceExists",
    "FlavorExists",
    "FlavorIdExists",
    "FlavorAccessExists",
    "InvalidSharedStorage",
    "InvalidLocalStorage",
    "StorageError",
    "MigrationError",
    "MigrationPreCheckError",
    "MigrationSchedulerRPCError",
    "MalformedRequestBody",
    "ConfigNotFound",
    "PasteAppNotFound",
    "CannotResizeToSameFlavor",
    "ResizeError",
    "CannotResizeDisk",
    "FlavorMemoryTooSmall",
    "FlavorDiskTooSmall",
    "FlavorDiskSmallerThanImage",
    "FlavorDiskSmallerThanMinDisk",
    "VolumeSmallerThanMinDisk",
    "BootFromVolumeRequiredForZeroDiskFlavor",
    "NoValidHost",
    "RequestFilterFailed",
    "InvalidRoutedNetworkConfiguration",
    "MaxRetriesExceeded",
    "QuotaError",
    "TooManyInstances",
    "FloatingIpLimitExceeded",
    "MetadataLimitExceeded",
    "OnsetFileLimitExceeded",
    "OnsetFilePathLimitExceeded",
    "OnsetFileContentLimitExceeded",
    "KeypairLimitExceeded",
    "SecurityGroupLimitExceeded",
    "PortLimitExceeded",
    "AggregateNotFound",
    "AggregateNameExists",
    "AggregateHostNotFound",
    "AggregateMetadataNotFound",
    "AggregateHostExists",
    "InstancePasswordSetFailed",
    "InstanceNotFound",
    "InstanceInfoCacheNotFound",
    "MarkerNotFound",
    "CouldNotFetchImage",
    "CouldNotUploadImage",
    "TaskAlreadyRunning",
    "TaskNotRunning",
    "InstanceIsLocked",
    "ConfigDriveInvalidValue",
    "ConfigDriveUnsupportedFormat",
    "ConfigDriveMountFailed",
    "ConfigDriveUnknownFormat",
    "ConfigDriveNotFound",
    "InterfaceAttachFailed",
    "InterfaceAttachFailedNoNetwork",
    "InterfaceAttachPciClaimFailed",
    "InterfaceAttachResourceAllocationFailed",
    "InterfaceDetachFailed",
    "InstanceUserDataMalformed",
    "InstanceUpdateConflict",
    "UnknownInstanceUpdateConflict",
    "UnexpectedTaskStateError",
    "UnexpectedDeletingTaskStateError",
    "InstanceActionNotFound",
    "InstanceActionEventNotFound",
    "InstanceEvacuateNotSupported",
    "DBNotAllowed",
    "UnsupportedVirtType",
    "UnsupportedHardware",
    "UnsupportedRescueBus",
    "UnsupportedRescueDevice",
    "UnsupportedRescueImage",
    "Base64Exception",
    "BuildAbortException",
    "RescheduledException",
    "ShadowTableExists",
    "InstanceFaultRollback",
    "OrphanedObjectError",
    "ObjectActionError",
    "InstanceGroupNotFound",
    "InstanceGroupIdExists",
    "InstanceGroupSaveException",
    "ResourceMonitorError",
    "PciDeviceWrongAddressFormat",
    "PciDeviceInvalidDeviceName",
    "PciDeviceNotFoundById",
    "PciDeviceNotFound",
    "PciDeviceInvalidStatus",
    "PciDeviceVFInvalidStatus",
    "PciDevicePFInvalidStatus",
    "PciDeviceInvalidOwner",
    "PciDeviceRequestFailed",
    "PciDevicePoolEmpty",
    "PciInvalidAlias",
    "PciRequestAliasNotDefined",
    "PciConfigInvalidWhitelist",
    "PciRequestFromVIFNotFound",
    "InternalError",
    "PciDeviceDetachFailed",
    "PciDeviceUnsupportedHypervisor",
    "KeyManagerError",
    "VolumesNotRemoved",
    "VolumeRebaseFailed",
    "InvalidVideoMode",
    "RngDeviceNotExist",
    "RequestedVRamTooHigh",
    "SecurityProxyNegotiationFailed",
    "RFBAuthHandshakeFailed",
    "RFBAuthNoAvailableScheme",
    "InvalidWatchdogAction",
    "LiveMigrationNotSubmitted",
    "SelectionObjectsWithOldRPCVersionNotSupported",
    "LiveMigrationURINotAvailable",
    "UnshelveException",
    "MismatchVolumeAZException",
    "UnshelveInstanceInvalidState",
    "ImageVCPULimitsRangeExceeded",
    "ImageVCPUTopologyRangeExceeded",
    "ImageVCPULimitsRangeImpossible",
    "InvalidArchitectureName",
    "ImageNUMATopologyIncomplete",
    "ImageNUMATopologyForbidden",
    "ImageNUMATopologyRebuildConflict",
    "ImagePCINUMAPolicyForbidden",
    "ImageNUMATopologyAsymmetric",
    "ImageNUMATopologyCPUOutOfRange",
    "ImageNUMATopologyCPUDuplicates",
    "ImageNUMATopologyCPUsUnassigned",
    "ImageNUMATopologyMemoryOutOfRange",
    "InvalidHostname",
    "NumaTopologyNotFound",
    "MigrationContextNotFound",
    "SocketPortRangeExhaustedException",
    "SocketPortInUseException",
    "ImageSerialPortNumberInvalid",
    "ImageSerialPortNumberExceedFlavorValue",
    "SerialPortNumberLimitExceeded",
    "InvalidImageConfigDrive",
    "InvalidHypervisorVirtType",
    "InvalidMachineType",
    "InvalidMachineTypeUpdate",
    "UnsupportedMachineType",
    "InvalidVirtualMachineMode",
    "InvalidToken",
    "TokenInUse",
    "InvalidConnectionInfo",
    "InstanceQuiesceNotSupported",
    "InstanceAgentNotEnabled",
    "QemuGuestAgentNotEnabled",
    "SetAdminPasswdNotSupported",
    "MemoryPageSizeInvalid",
    "MemoryPageSizeForbidden",
    "MemoryPageSizeNotSupported",
    "CPUPinningInvalid",
    "CPUUnpinningInvalid",
    "CPUPinningUnknown",
    "CPUUnpinningUnknown",
    "ImageCPUPinningForbidden",
    "ImageCPUThreadPolicyForbidden",
    "ImagePMUConflict",
    "UnsupportedPolicyException",
    "CellMappingNotFound",
    "NUMATopologyUnsupported",
    "MemoryPagesUnsupported",
    "InvalidImageFormat",
    "UnsupportedImageModel",
    "HostMappingNotFound",
    "HostMappingExists",
    "RealtimeConfigurationInvalid",
    "CPUThreadPolicyConfigurationInvalid",
    "RequestSpecNotFound",
    "UEFINotSupported",
    "SecureBootNotSupported",
    "TriggerCrashDumpNotSupported",
    "UnsupportedHostCPUControlPolicy",
    "LibguestfsCannotReadKernel",
    "RealtimeMaskNotFoundOrInvalid",
    "OsInfoNotFound",
    "BuildRequestNotFound",
    "AttachInterfaceNotSupported",
    "AttachInterfaceWithQoSPolicyNotSupported",
    "NetworksWithQoSPolicyNotSupported",
    "CreateWithPortResourceRequestOldVersion",
    "InvalidReservedMemoryPagesOption",
    "ResourceProviderInUse",
    "ResourceProviderRetrievalFailed",
    "ResourceProviderAggregateRetrievalFailed",
    "ResourceProviderTraitRetrievalFailed",
    "ResourceProviderCreationFailed",
    "ResourceProviderDeletionFailed",
    "ResourceProviderUpdateFailed",
    "ResourceProviderNotFound",
    "ResourceProviderSyncFailed",
    "PlacementAPIConnectFailure",
    "PlacementAPIConflict",
    "ResourceProviderUpdateConflict",
    "InvalidResourceClass",
    "InvalidInventory",
    "InventoryInUse",
    "UsagesRetrievalFailed",
    "NotSupportedWithOption",
    "Unauthorized",
    "NeutronAdminCredentialConfigurationInvalid",
    "InvalidEmulatorThreadsPolicy",
    "InvalidCPUAllocationPolicy",
    "InvalidCPUThreadAllocationPolicy",
    "BadRequirementEmulatorThreadsPolicy",
    "InvalidNetworkNUMAAffinity",
    "InvalidPCINUMAAffinity",
    "PowerVMAPIFailed",
    "TraitRetrievalFailed",
    "TraitCreationFailed",
    "CannotMigrateToSameHost",
    "VirtDriverNotReady",
    "InvalidPeerList",
    "InstanceDiskMappingFailed",
    "NewMgmtMappingNotFoundException",
    "NoDiskDiscoveryException",
    "UniqueDiskDiscoveryException",
    "DeviceDeletionException",
    "OptRequiredIfOtherOptValue",
    "AllocationCreateFailed",
    "AllocationUpdateFailed",
    "AllocationMoveFailed",
    "AllocationDeleteFailed",
    "TooManyComputesForHost",
    "CertificateValidationFailed",
    "InstanceRescueFailure",
    "InstanceUnRescueFailure",
    "IronicAPIVersionNotAvailable",
    "ZVMDriverException",
    "ZVMConnectorError",
    "NoResourceClass",
    "ResourceProviderAllocationRetrievalFailed",
    "ConsumerAllocationRetrievalFailed",
    "ReshapeFailed",
    "ReshapeNeeded",
    "FlavorImageConflict",
    "MissingDomainCapabilityFeatureException",
    "HealPortAllocationException",
    "MoreThanOneResourceProviderToHealFrom",
    "NoResourceProviderToHealFrom",
    "UnableToQueryPorts",
    "UnableToUpdatePorts",
    "UnableToRollbackPortUpdates",
    "AssignedResourceNotFound",
    "PMEMNamespaceConfigInvalid",
    "GetPMEMNamespacesFailed",
    "VPMEMCleanupFailed",
    "RequestGroupSuffixConflict",
    "AmbiguousResourceProviderForPCIRequest",
    "UnexpectedResourceProviderNameForPCIRequest",
    "DeviceProfileError",
    "AcceleratorRequestOpFailed",
    "AcceleratorRequestBindingFailed",
    "InvalidLibvirtGPUConfig",
    "RequiredMixedInstancePolicy",
    "RequiredMixedOrRealtimeCPUMask",
    "MixedInstanceNotSupportByComputeService",
    "InvalidMixedInstanceDedicatedMask",
    "ProviderConfigException",
]

# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' placement/exception.py
PLACEMENT_EXCEPTIONS = [
    "NotFound",
    "Exists",
    "InvalidInventory",
    "CannotDeleteParentResourceProvider",
    "ConcurrentUpdateDetected",
    "ResourceProviderConcurrentUpdateDetected",
    "ResourceProviderNotFound",
    "InvalidAllocationCapacityExceeded",
    "InvalidAllocationConstraintsViolated",
    "InvalidInventoryCapacity",
    "InvalidInventoryCapacityReservedCanBeTotal",
    "InventoryInUse",
    "InventoryWithResourceClassNotFound",
    "MaxDBRetriesExceeded",
    "ObjectActionError",
    "PolicyNotAuthorized",
    "ResourceClassCannotDeleteStandard",
    "ResourceClassCannotUpdateStandard",
    "ResourceClassExists",
    "ResourceClassInUse",
    "ResourceClassNotFound",
    "ResourceProviderInUse",
    "TraitCannotDeleteStandard",
    "TraitExists",
    "TraitInUse",
    "TraitNotFound",
    "ProjectNotFound",
    "ProjectExists",
    "UserNotFound",
    "UserExists",
    "ConsumerNotFound",
    "ConsumerExists",
    "ConsumerTypeNotFound",
    "ConsumerTypeExists",
]

# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p'  ./os_vif/exception.py
# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p'  ./vif_plug_ovs/exception.py
_OS_VIF_EXCEPTIONS = [
    "ExceptionBase",
    "LibraryNotInitialized",
    "NoMatchingPlugin",
    "NoMatchingPortProfileClass",
    "NoSupportedPortProfileVersion",
    "NoMatchingVIFClass",
    "NoSupportedVIFVersion",
    "PlugException",
    "UnplugException",
    "NetworkMissingPhysicalNetwork",
    "NetworkInterfaceNotFound",
    "NetworkInterfaceTypeNotDefined",
    "ExternalImport",
    "NotImplementedForOS",
    "AgentError",
    "MissingPortProfile",
    "WrongPortProfile",
    "RepresentorNotFound",
    "PciDeviceNotFoundById",
]
OS_VIF_EXCEPTIONS = [f"os_vif.exception.{exc}"
                     for exc in _OS_VIF_EXCEPTIONS]



from .common import (
    intercept_exception,
    OpsUtils,
    CheckItemsBase,
    YRequirementTypeBase,
    YRequirementTypeWithOpsBase,
    PackageCheckItemsBase,
    ServiceCheckItemsBase,
)

__all__ = [
    intercept_exception.__name__,
    OpsUtils.__name__,
    CheckItemsBase.__name__,
    YRequirementTypeBase.__name__,
    YRequirementTypeWithOpsBase.__name__,
    PackageCheckItemsBase.__name__,
    ServiceCheckItemsBase.__name__,
    ]



import os
import re
from datetime import datetime, timedelta

from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers import CLIHelper
from hotsos.core.log import log
from hotsos.core.plugins.juju.common import JujuChecks
from hotsos.core.search import (
    FileSearcher,
    SearchDef,
    SearchConstraintSearchSince,
)
from hotsos.core.utils import sorted_dict
from hotsos.core.search import CommonTimestampMatcher
from hotsos.core.plugintools import (
    summary_entry,
    get_min_available_entry_index,
)


def _tally_result(app_name, events, result, tally_key):
    if app_name not in events:
        events[app_name] = {}

    tag = result.tag.lower()
    if tag not in events[app_name]:
        events[app_name][tag] = {}

    origin = result.get(3)
    origin_child = origin.rpartition('.')[2]
    if origin_child not in events[app_name][tag]:
        events[app_name][tag][origin_child] = {}

    mod = result.get(4)
    if mod not in events[app_name][tag][origin_child]:
        events[app_name][tag][origin_child][mod] = {}

    if tally_key not in events[app_name][tag][origin_child][mod]:
        events[app_name][tag][origin_child][mod][tally_key] = 1
    else:
        events[app_name][tag][origin_child][mod][tally_key] += 1


def _get_app_name(searchobj, source_id):
    path = searchobj.resolve_source_id(source_id)
    app_name = re.search(r".+/unit-(\S+).log.*", path).group(1)
    return app_name


def _should_skip_log(now, then):
    # Since juju logs files don't typically get logrotated they
    # may contain a large history of logs so we have to do this to
    # ensure we don't get too much.
    days = 1
    if HotSOSConfig.use_all_logs:
        days = HotSOSConfig.max_logrotate_depth

    return then < now - timedelta(days=days)


def _init_searchobj():
    c = SearchConstraintSearchSince(ts_matcher_cls=CommonTimestampMatcher)
    searchobj = FileSearcher(constraint=c)
    path = os.path.join(HotSOSConfig.data_root, 'var/log/juju/unit-*.log')
    ts_expr = r"^([\d-]+)\s+([\d:]+)"
    for msg in ['ERROR', 'WARNING']:
        expr = rf'{ts_expr} {msg} (\S+) (\S+):\d+ '
        tag = msg
        hint = msg
        searchobj.add(SearchDef(expr, tag=tag, hint=hint), path)

    return searchobj


def get_error_and_warnings():
    """ Create a tally of log errors and warnings for each unit. """
    log.debug("searching unit logs for errors and warnings")
    searchobj = _init_searchobj()

    results = searchobj.run()
    log.debug("fetching unit log results")
    events = {}
    date_format = CommonTimestampMatcher.DEFAULT_DATETIME_FORMAT
    now = CLIHelper().date(format=f"+{date_format}")
    now = datetime.strptime(now, date_format)

    for tag in ['WARNING', 'ERROR']:
        for result in results.find_by_tag(tag):
            ts_date = result.get(1)
            if HotSOSConfig.event_tally_granularity == 'time':
                ts_time = result.get(2)
                # use hours and minutes only
                ts_time = re.compile(r'(\d+:\d+).+').search(ts_time)[1]
                key = f"{ts_date}_{ts_time}"
            else:
                ts_time = '00:00:00'
                key = ts_date

            then = datetime.strptime(f"{ts_date} {ts_time}", date_format)
            if _should_skip_log(now, then):
                continue

            app_name = _get_app_name(searchobj, result.source_id)
            _tally_result(app_name, events, result, key)

    # ensure consistent ordering of results
    for tag, units in events.items():
        for unit, keys in units.items():
            units[unit] = sorted_dict(keys)

        events[tag] = sorted_dict(units)

    return sorted_dict(events)


class JujuSummary(JujuChecks):
    """ Implementation of Juju summary. """
    summary_part_index = 0

    @summary_entry('machine', get_min_available_entry_index())
    def summary_machine(self):
        if self.machine:
            return self.machine.id

        return "unknown"

    @summary_entry('units', get_min_available_entry_index() + 1)
    def summary_units(self):
        if not self.units:
            return None

        unit_info = {}
        loginfo = get_error_and_warnings()
        for u in self.units.values():
            name, _, ver = u.name.rpartition('-')
            u_name = f"{name}/{ver}"
            unit_info[u_name] = {}
            if u.repo_info:
                c_name = u.charm_name
                if c_name:
                    unit_info[u_name]['charm'] = {'name': c_name}
                    sha1 = u.repo_info.get('commit')
                    unit_info[u_name]['charm']['repo-info'] = sha1
                    if c_name in self.charms:
                        unit_info[u_name]['charm']['version'] = \
                            self.charms[c_name].version

            if u.name in loginfo:
                unit_info[u_name]['logs'] = loginfo[u.name]

        return unit_info or None



# flake8: noqa
# pylint: disable=C0301

# Neutron exception definitions are distributed across multiple modules/files
# and projects.
#
# neutron:
# find -name exception\*.py| xargs -l sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p'
#
# neutron-lib:
# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' ./neutron_lib/exceptions/*.py
#
# neutron extras
# sed -rn 's/\S+ (\S+)\(RuntimeError\):.*/    "\1",/p' ./neutron/privileged/agent/linux/ip_lib.py
NEUTRON_EXCEPTIONS = [
    # repo:neutron
    "OVSDBPortError",
    "SriovUnsupportedNetworkType",
    "SriovNicError",
    "InvalidDeviceError",
    "InvalidPciSlotError",
    "MechanismDriverError",
    "ExtensionDriverError",
    "ExtensionDriverNotFound",
    "UnknownNetworkType",
    "OVSFWPortNotFound",
    "OVSFWTagNotFound",
    "OVSFWPortNotHandled",
    "OvsdbSslConfigNotFound",
    "OvsdbSslRequiredOptError",
    "RevisionConflict",
    "UnknownResourceType",
    "StandardAttributeIDNotFound",
    "HashRingIsEmpty",
    "InvalidSubnetRequestType",
    "AddressCalculationFailure",
    "InvalidAddressType",
    "IpAddressAlreadyAllocated",
    "InvalidIpForSubnet",
    "InvalidAddressRequest",
    "InvalidSubnetRequest",
    "IPAddressChangeNotAllowed",
    "AllocationOnAutoAddressSubnet",
    "IpAddressGenerationFailure",
    "IpAddressGenerationFailureAllSubnets",
    "IpAddressGenerationFailureNoMatchingSubnet",
    "IPAllocationFailed",
    "IpamValueInvalid",
    "DeferIpam",
    "SegmentNotFound",
    "NoUpdateSubnetWhenMultipleSegmentsOnNetwork",
    "SubnetsNotAllAssociatedWithSegments",
    "SubnetCantAssociateToDynamicSegment",
    "SubnetSegmentAssociationChangeNotAllowed",
    "NetworkIdsDontMatch",
    "HostNotConnectedToAnySegment",
    "HostNotCompatibleWithFixedIps",
    "SegmentInUse",
    "FixedIpsSubnetsNotOnSameSegment",
    "AutoAllocationFailure",
    "DefaultExternalNetworkExists",
    "UnknownProvisioningError",
    "ConntrackHelperNotFound",
    "ConntrackHelperNotAllowed",
    "InvalidProtocolForHelper",
    "RouterGatewayInUseByNDPProxy",
    "RouterInterfaceInUseByNDPProxy",
    "AddressScopeConflict",
    "RouterGatewayNotValid",
    "RouterNDPProxyNotEnable",
    "PortUnreachableRouter",
    "InvalidAddress",
    "RouterIPv6GatewayInUse",
    "NDPProxyNotFound",
    "LogResourceNotFound",
    "InvalidLogResourceType",
    "LoggingTypeNotSupported",
    "TargetResourceNotFound",
    "ResourceNotFound",
    "InvalidResourceConstraint",
    "LogapiDriverException",
    "CookieNotFound",
    "ValidatedMethodNotFound",
    "ResourceIdNotSpecified",
    "RouterNotEnabledSnat",
    "EventsDisabled",
    "RouterGatewayNotSet",
    "PortForwardingNotFound",
    "PortForwardingNotSupportFilterField",
    "PortHasPortForwarding",
    "FipInUseByPortForwarding",
    "PortHasBindingFloatingIP",
    "MechanismDriverNotFound",
    "TrunkBridgeNotFound",
    "ParentPortNotFound",
    "TrunkPortInUse",
    "TrunkNotFound",
    "SubPortNotFound",
    "DuplicateSubPort",
    "ParentPortInUse",
    "SubPortMtuGreaterThanTrunkPortMtu",
    "PortInUseAsTrunkParent",
    "PortInUseAsSubPort",
    "TrunkInUse",
    "TrunkDisabled",
    "TrunkInErrorState",
    "IncompatibleTrunkPluginConfiguration",
    "IncompatibleDriverSegmentationTypes",
    "SegmentationTypeValidatorNotFound",
    "TrunkPluginDriverConflict",
    "SubPortBindingError",
    "CallbackWrongResourceType",
    "CallbackNotFound",
    "CallbacksMaxLimitReached",
    "NoAgentDbMixinImplemented",
    # repo: neutron (extras)
    "NetworkNamespaceNotFound",
    "NetworkInterfaceNotFound",
    "InterfaceOperationNotSupported",
    "InvalidArgument",
    "IpAddressAlreadyExists",
    "InterfaceAlreadyExists",
    # repo:neutron-lib
    "AddressGroupNotFound",
    "AddressGroupInUse",
    "AddressesNotFound",
    "AddressesAlreadyExist",
    "AddressScopeNotFound",
    "AddressScopeInUse",
    "AddressScopeUpdateError",
    "NetworkAddressScopeAffinityError",
    "AgentNotFound",
    "AgentNotFoundByTypeHost",
    "MultipleAgentFoundByTypeHost",
    "AllowedAddressPairsMissingIP",
    "AddressPairAndPortSecurityRequired",
    "DuplicateAddressPairInRequest",
    "AllowedAddressPairExhausted",
    "AvailabilityZoneNotFound",
    "InvalidDHCPAgent",
    "NetworkHostedByDHCPAgent",
    "NetworkNotHostedByDhcpAgent",
    "DNSDomainNotFound",
    "DuplicateRecordSet",
    "ExternalDNSDriverNotFound",
    "InvalidPTRZoneConfiguration",
    "DVRMacAddressNotFound",
    "ExternalNetworkInUse",
    "InvalidRoutes",
    "RouterInterfaceInUseByRoute",
    "RoutesExhausted",
    "FirewallGroupNotFound",
    "FirewallGroupInUse",
    "FirewallGroupInPendingState",
    "FirewallGroupPortInvalid",
    "FirewallGroupPortInvalidProject",
    "FirewallGroupPortInUse",
    "FirewallPolicyNotFound",
    "FirewallPolicyInUse",
    "FirewallPolicyConflict",
    "FirewallRuleSharingConflict",
    "FirewallPolicySharingConflict",
    "FirewallRuleNotFound",
    "FirewallRuleInUse",
    "FirewallRuleNotAssociatedWithPolicy",
    "FirewallRuleInvalidProtocol",
    "FirewallRuleInvalidAction",
    "FirewallRuleInvalidICMPParameter",
    "FirewallRuleWithPortWithoutProtocolInvalid",
    "FirewallRuleInvalidPortValue",
    "FirewallRuleInfoMissing",
    "FirewallIpAddressConflict",
    "FirewallInternalDriverError",
    "FirewallRuleConflict",
    "FirewallRuleAlreadyAssociated",
    "FirewallGroupCannotRemoveDefault",
    "FirewallGroupCannotUpdateDefault",
    "FirewallGroupDefaultAlreadyExists",
    "FlavorNotFound",
    "FlavorInUse",
    "ServiceProfileNotFound",
    "ServiceProfileInUse",
    "FlavorServiceProfileBindingExists",
    "FlavorServiceProfileBindingNotFound",
    "ServiceProfileDriverNotFound",
    "ServiceProfileEmpty",
    "FlavorDisabled",
    "ServiceProfileDisabled",
    "NeutronException",
    "BadRequest",
    "NotFound",
    "Conflict",
    "NotAuthorized",
    "ServiceUnavailable",
    "AdminRequired",
    "ObjectNotFound",
    "NetworkNotFound",
    "SubnetNotFound",
    "PortNotFound",
    "PortNotFoundOnNetwork",
    "DeviceNotFoundError",
    "InUse",
    "NetworkInUse",
    "SubnetInUse",
    "SubnetPoolInUse",
    "PortInUse",
    "ServicePortInUse",
    "PortBound",
    "PortBoundNUMAAffinityPolicy",
    "MacAddressInUse",
    "InvalidIpForNetwork",
    "InvalidIpForSubnet",
    "IpAddressInUse",
    "VlanIdInUse",
    "TunnelIdInUse",
    "ResourceExhausted",
    "NoNetworkAvailable",
    "SubnetMismatchForPort",
    "Invalid",
    "InvalidInput",
    "IpAddressGenerationFailure",
    "PreexistingDeviceFailure",
    "OverQuota",
    "InvalidContentType",
    "ExternalIpAddressExhausted",
    "InvalidConfigurationOption",
    "NetworkTunnelRangeError",
    "PolicyInitError",
    "PolicyCheckError",
    "MultipleExceptions",
    "HostMacAddressGenerationFailure",
    "NetworkMacAddressGenerationFailure",
    "InvalidServiceType",
    "NetworkVlanRangeError",
    "PhysicalNetworkNameError",
    "TenantIdProjectIdFilterConflict",
    "SubnetPoolNotFound",
    "StateInvalid",
    "DhcpPortInUse",
    "HostRoutesExhausted",
    "DNSNameServersExhausted",
    "FlatNetworkInUse",
    "NoNetworkFoundInMaximumAllowedAttempts",
    "MalformedRequestBody",
    "InvalidAllocationPool",
    "UnsupportedPortDeviceOwner",
    "OverlappingAllocationPools",
    "OutOfBoundsAllocationPool",
    "BridgeDoesNotExist",
    "QuotaResourceUnknown",
    "QuotaMissingTenant",
    "InvalidQuotaValue",
    "InvalidSharedSetting",
    "ExtensionsNotFound",
    "GatewayConflictWithAllocationPools",
    "GatewayIpInUse",
    "NetworkVxlanPortRangeError",
    "VxlanNetworkUnsupported",
    "DuplicatedExtension",
    "DriverCallError",
    "DeviceIDNotOwnedByTenant",
    "InvalidCIDR",
    "FailToDropPrivilegesExit",
    "NetworkIdOrRouterIdRequiredError",
    "EmptySubnetPoolPrefixList",
    "PrefixVersionMismatch",
    "UnsupportedMinSubnetPoolPrefix",
    "IllegalSubnetPoolPrefixBounds",
    "IllegalSubnetPoolPrefixUpdate",
    "SubnetAllocationError",
    "AddressScopePrefixConflict",
    "IllegalSubnetPoolAssociationToAddressScope",
    "IllegalSubnetPoolIpVersionAssociationToAddressScope",
    "IllegalSubnetPoolUpdate",
    "MinPrefixSubnetAllocationError",
    "MaxPrefixSubnetAllocationError",
    "SubnetPoolDeleteError",
    "SubnetPoolQuotaExceeded",
    "NetworkSubnetPoolAffinityError",
    "ObjectActionError",
    "CTZoneExhaustedError",
    "TenantQuotaNotFound",
    "MultipleFilterIDForIPFound",
    "FilterIDForIPNotFound",
    "FailedToAddQdiscToDevice",
    "PortBindingNotFound",
    "PortBindingAlreadyActive",
    "PortBindingAlreadyExists",
    "PortBindingError",
    "ProcessExecutionError",
    "InvalidSubnetServiceType",
    "InvalidInputSubnetServiceType",
    "MaxVRIDAllocationTriesReached",
    "NoVRIDAvailable",
    "HANetworkConcurrentDeletion",
    "HANetworkCIDRNotValid",
    "HAMaximumAgentsNumberNotValid",
    "RouterNotFound",
    "RouterInUse",
    "RouterInterfaceNotFound",
    "RouterInterfaceNotFoundForSubnet",
    "RouterInterfaceInUseByFloatingIP",
    "FloatingIPNotFound",
    "ExternalGatewayForFloatingIPNotFound",
    "FloatingIPPortAlreadyAssociated",
    "RouterExternalGatewayInUseByFloatingIp",
    "RouterInterfaceAttachmentConflict",
    "RouterNotCompatibleWithAgent",
    "RouterNotFoundInRouterFactory",
    "FloatingIpSetupException",
    "AbortSyncRouters",
    "IpTablesApplyException",
    "MeteringLabelNotFound",
    "DuplicateMeteringRuleInPost",
    "MeteringLabelRuleNotFound",
    "MeteringLabelRuleOverlaps",
    "SegmentsSetInConjunctionWithProviders",
    "SegmentsContainDuplicateEntry",
    "NetworkSegmentRangeNetTypeNotSupported",
    "NetworkSegmentRangeNotFound",
    "NetworkSegmentRangeReferencedByProject",
    "NetworkSegmentRangeDefaultReadOnly",
    "NetworkSegmentRangeOverlaps",
    "PlacementEndpointNotFound",
    "PlacementResourceNotFound",
    "PlacementResourceProviderNotFound",
    "PlacementResourceProviderGenerationConflict",
    "PlacementInventoryNotFound",
    "PlacementInventoryUpdateConflict",
    "PlacementAggregateNotFound",
    "PlacementTraitNotFound",
    "PlacementResourceClassNotFound",
    "PlacementAPIVersionIncorrect",
    "PlacementResourceProviderNameNotUnique",
    "PlacementClientError",
    "UnknownResourceProvider",
    "AmbiguousResponsibilityForResourceProvider",
    "PlacementAllocationGenerationConflict",
    "PlacementAllocationRemoved",
    "PlacementAllocationRpNotExists",
    "PortSecurityPortHasSecurityGroup",
    "PortSecurityAndIPRequiredForSecurityGroups",
    "QosPolicyNotFound",
    "QosRuleNotFound",
    "QoSPolicyDefaultAlreadyExists",
    "PortQosBindingNotFound",
    "PortQosBindingError",
    "NetworkQosBindingNotFound",
    "FloatingIPQosBindingNotFound",
    "QosPolicyInUse",
    "FloatingIPQosBindingError",
    "NetworkQosBindingError",
    "QosRuleNotSupported",
    "QoSRuleParameterConflict",
    "QoSRulesConflict",
    "PolicyRemoveAuthorizationError",
    "TcLibQdiscTypeError",
    "TcLibQdiscNeededArguments",
    "RouterQosBindingNotFound",
    "RouterQosBindingError",
    "QosPlacementAllocationConflict",
    "VlanTransparencyDriverError",
    "VPNServiceNotFound",
    "IPsecSiteConnectionNotFound",
    "IPsecSiteConnectionDpdIntervalValueError",
    "IPsecSiteConnectionMtuError",
    "IPsecSiteConnectionPeerCidrError",
    "IKEPolicyNotFound",
    "IPsecPolicyNotFound",
    "IKEPolicyInUse",
    "VPNServiceInUse",
    "SubnetInUseByVPNService",
    "SubnetInUseByEndpointGroup",
    "SubnetInUseByIPsecSiteConnection",
    "VPNStateInvalidToUpdate",
    "IPsecPolicyInUse",
    "DeviceDriverImportError",
    "SubnetIsNotConnectedToRouter",
    "RouterIsNotExternal",
    "VPNPeerAddressNotResolved",
    "ExternalNetworkHasNoSubnet",
    "VPNEndpointGroupNotFound",
    "InvalidEndpointInEndpointGroup",
    "MissingEndpointForEndpointGroup",
    "NonExistingSubnetInEndpointGroup",
    "MixedIPVersionsForIPSecEndpoints",
    "MixedIPVersionsForPeerCidrs",
    "MixedIPVersionsForIPSecConnection",
    "InvalidEndpointGroup",
    "WrongEndpointGroupType",
    "PeerCidrsInvalid",
    "MissingPeerCidrs",
    "MissingRequiredEndpointGroup",
    "EndpointGroupInUse",
    "FlavorsPluginNotLoaded",
    "NoProviderFoundForFlavor",
    "IpsecValidationFailure",
    "IkeValidationFailure",
    "CsrInternalError",
    "CsrValidationFailure",
]

# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' ./neutronclient/common/exceptions.py
_NEUTRONCLIENT_EXCEPTIONS = [
    "NeutronException",
    "NeutronClientException",
    "BadRequest",
    "Unauthorized",
    "Forbidden",
    "NotFound",
    "Conflict",
    "InternalServerError",
    "ServiceUnavailable",
    "NetworkNotFoundClient",
    "PortNotFoundClient",
    "StateInvalidClient",
    "NetworkInUseClient",
    "PortInUseClient",
    "IpAddressInUseClient",
    "IpAddressAlreadyAllocatedClient",
    "InvalidIpForNetworkClient",
    "InvalidIpForSubnetClient",
    "OverQuotaClient",
    "IpAddressGenerationFailureClient",
    "MacAddressInUseClient",
    "HostNotCompatibleWithFixedIpsClient",
    "ExternalIpAddressExhaustedClient",
    "NoAuthURLProvided",
    "EndpointNotFound",
    "EndpointTypeNotFound",
    "AmbiguousEndpoints",
    "RequestURITooLong",
    "ConnectionFailed",
    "SslCertificateValidationError",
    "MalformedResponseBody",
    "InvalidContentType",
    "NeutronCLIError",
    "CommandError",
    "UnsupportedVersion",
    "NeutronClientNoUniqueMatch",
]
NEUTRONCLIENT_EXCEPTIONS = [f"neutronclient.common.exceptions.{exc}"
                            for exc in _NEUTRONCLIENT_EXCEPTIONS]

# Including this as a dep of (at least) Neutron
# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' ovsdbapp/exceptions.py
OVSDBAPP_EXCEPTIONS = [
    "OvsdbAppException",
    "TimeoutException",
    "OvsdbConnectionUnavailable",
]



import abc
import re
import subprocess
from dataclasses import dataclass

from hotsos.core.factory import FactoryBase
from hotsos.core.host_helpers.cli import CLIHelper
from hotsos.core.log import log
from hotsos.core.utils import sorted_dict

lower_bound_ops = ["gt", "ge", "eq"]  # ops that define a lower bound
upper_bound_ops = ["lt", "le", "eq"]  # ops that define an upper bound
equal_compr_ops = ["eq", "ge", "le"]  # ops that compare for equality


class DPKGBadVersionSyntax(Exception):
    """ Exception raised when an invalid version is provided in a check. """


class DPKGVersion():
    """ Helper for querying and comparing dpkg packaging versions. """
    def __init__(self, a):
        self.a = str(a)

    def _compare_impl(self, op, b):
        try:
            output = subprocess.check_output(['dpkg', '--compare-versions',
                                              self.a, op, b],
                                             stderr=subprocess.STDOUT)
            if re.search(b"dpkg: warning: version.*has bad syntax:.*", output):
                raise DPKGBadVersionSyntax(output)
        except subprocess.CalledProcessError as se:
            if se.returncode == 1:
                return False

            raise se

        return True

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.a

    def __eq__(self, b):
        return self._compare_impl('eq', str(b))

    def __lt__(self, b):
        return not self._compare_impl('ge', str(b))

    def __gt__(self, b):
        return not self._compare_impl('le', str(b))

    def __le__(self, b):
        return self._compare_impl('le', str(b))

    def __ge__(self, b):
        return self._compare_impl('ge', str(b))

    @staticmethod
    def add_version_criteria_bounds(version_criteria):
        """Add upper/lower bounds to criteria, where needed

        @param version_criteria: List of version ranges
        @return: List of version ranges with upper/lower bounds
        """
        # Step 5: Add the implicit upper/lower bounds where needed
        for idx, version_criterion in enumerate(version_criteria):
            log.debug("\tchecking criterion %s", str(version_criterion))

            has_lower_bound = any(x in lower_bound_ops
                                  for x in version_criterion)
            has_upper_bound = any(x in upper_bound_ops
                                  for x in version_criterion)
            is_the_last_item = idx == (len(version_criteria) - 1)
            is_the_first_item = idx == 0

            log.debug("\t\tcriterion %s has lower bound?"
                      "%s has upper bound? %s", str(version_criterion),
                      has_lower_bound, has_upper_bound)

            if not has_upper_bound and not is_the_last_item:
                op = "le"  # default
                next_criterion = version_criteria[idx + 1]
                next_op, next_val = list(next_criterion.items())[0]
                # If the next criterion op compares for equality, then the
                # implicit op added to this criterion should not compare for
                # equality.
                if next_op in equal_compr_ops:
                    op = "lt"
                log.debug("\t\tadding implicit upper bound %s:%s to %s", op,
                          next_val, version_criterion)
                version_criterion[op] = next_val
            elif not has_lower_bound and not is_the_first_item:
                op = "ge"  # default
                prev_criterion = version_criteria[idx - 1]
                prev_op, prev_val = list(prev_criterion.items())[-1]
                # If the previous criterion op compares for equality, then the
                # implicit op added to this criterion should not compare for
                # equality.
                if prev_op in equal_compr_ops:
                    op = "gt"
                log.debug("\t\tadding implicit lower bound %s:%s to %s", op,
                          prev_val, version_criterion)
                version_criterion[op] = prev_val

            # Re-sort and overwrite the criterion
            version_criteria[idx] = dict(
                sorted(version_criterion.items(),
                       key=lambda a: a[1]))

    @staticmethod
    def normalize_version_criteria(version_criteria):
        """Normalize all the criterions in a criteria.

        Normalization does the following:
        - removes empty criteria
        - replaces old ops with the new ones
        - sorts each criterion(ascending) and criteria(descending)
        - adds upper/lower bounds to criteria, where needed

        @param version_criteria: List of version ranges to normalize
        @return: Normalized list of version ranges
        """

        # Step 0: Ensure that all version values are DPKGVersion type
        for idx, version_criterion in enumerate(version_criteria):
            for k, v in version_criterion.items():
                version_criterion.update({k: DPKGVersion(v)})

        # Step 1: Remove empty criteria
        version_criteria = [x for x in version_criteria if len(x) > 0]

        # Step 2: Replace legacy ops with the new ones
        legacy_ops = {"min": "ge", "max": "le"}
        for idx, version_criterion in enumerate(version_criteria):
            for lop, nop in legacy_ops.items():
                if lop in version_criterion:
                    version_criterion[nop] = version_criterion[lop]
                    del version_criterion[lop]

        # Step 3: Sort each criterion in itself, so the smallest version
        # appears first
        for idx, version_criterion in enumerate(version_criteria):
            version_criterion = dict(sorted(version_criterion.items(),
                                     key=lambda a: a[1]))
            version_criteria[idx] = version_criterion

        # Step 4: Sort all criteria by the first element in the criterion
        version_criteria = sorted(version_criteria,
                                  key=lambda a: list(a.values())[0])

        # Step 5: Add the implicit upper/lower bounds where needed
        DPKGVersion.add_version_criteria_bounds(version_criteria)

        # Step 6: Sort by descending order so the largest version range
        # appears first
        version_criteria = sorted(version_criteria,
                                  key=lambda a: list(a.values())[0],
                                  reverse=True)

        log.debug("final criteria: %s", str(version_criteria))
        return version_criteria

    @staticmethod
    def is_version_within_ranges(version, version_criteria):
        """Check if pkg's version satisfies any criterion listed in
        the version_criteria.

        @param version: Version
        @param version_criteria: Criteria to check version against

        @return: True if ver(pkg) satisfies any criterion, false otherwise.
        """
        result = True

        # Supported operations for defining version ranges
        ops = {
            "eq": lambda lhs, rhs: lhs == DPKGVersion(rhs),
            "lt": lambda lhs, rhs: lhs < DPKGVersion(rhs),
            "le": lambda lhs, rhs: lhs <= DPKGVersion(rhs),
            "gt": lambda lhs, rhs: lhs > DPKGVersion(rhs),
            "ge": lambda lhs, rhs: lhs >= DPKGVersion(rhs),
            # Here, it's necessary to use lambda. It defers the
            # "ops" access to runtime, so ops["ge"] exists when
            # this is executed.
            # pylint: disable-next=unnecessary-lambda
            "min": lambda lhs, rhs: ops["ge"](lhs, rhs),
            # pylint: disable-next=unnecessary-lambda
            "max": lambda lhs, rhs: ops["le"](lhs, rhs),
        }

        version_criteria = DPKGVersion.normalize_version_criteria(
            version_criteria)

        for version_criterion in version_criteria:
            if not version_criterion.keys() <= ops.keys():
                raise NameError(
                    "Unrecognized comparison operator name(s):"
                    f"{version_criterion.keys()-ops.keys()}")
            # Each criterion is evaluated on its own
            # so if any of the criteria is true, then
            # the check is also true.
            for op_name, op_fn in ops.items():
                if op_name in version_criterion:
                    cversion = str(version_criterion[op_name])
                    # Check if the criterion is satisfied or not
                    if not op_fn(str(version), cversion):
                        break
            else:
                # Loop is not exited by a break which means
                # all ops in the criterion are satisfied.
                result = True
                log.debug("version criterion %s satisfied by %s",
                          version_criterion, version)
                # Break the outer loop
                break
            result = False

        log.debug("version %s within version ranges %s "
                  "(result=%s)", version, version_criteria, result)
        return result


class PackageHelperBase(abc.ABC):
    """ Base class for packaging helpers. """
    def get_version(self, pkg):
        """
        Return version of package.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def is_installed(self, pkg):
        """ Returns True if pkg is installed """

    @property
    def all_formatted(self):
        """
        Returns list of packages. Each entry in the list is an item
        looking like "<pkgname> <pkgver>".

        Converts dict to a list of strings of format '<name> <version>'.
        """

        _all = self.all
        if not _all:
            return []

        return [f"{e[0]} {e[1]}" for e in _all.items()]

    @property
    def all(self):
        """ Returns results matched for all expressions. """
        raise NotImplementedError

    @property
    def core(self):
        """ Returns results matched for core expressions. """
        raise NotImplementedError


class DockerImageHelper(PackageHelperBase):
    """ Helpers for analysing docker images. """
    def __init__(self, core_pkgs, other_pkgs=None):
        """
        @param core_pkgs:
        package names.
        @param other_pkg_exprs:
        """
        self.core_image_exprs = core_pkgs
        self.other_image_exprs = other_pkgs or []
        self._core_images = {}
        self._other_images = {}
        self._all_images = {}
        # The following expression muct match any package name.
        self._match_expr_template = \
            r"^(\S+/(\S+))\s+([\d\.]+)\s+(\S+)\s+.+"
        self.cli = CLIHelper()

    def is_installed(self, pkg):
        return pkg in self.all

    def _match_image(self, image, entry):
        expr = self._match_expr_template.format(image)
        ret = re.compile(expr).match(entry)
        if ret:
            return ret[1], ret[2], ret[3]

        return None, None, None

    def get_container_images(self):
        images = []
        for line in self.cli.docker_ps():
            ret = re.compile(r"^\S+\s+(\S+):(\S+)\s+.+").match(line)
            if not ret:
                continue

            images.append((ret[1], ret[2]))

        return images

    @property
    def _all(self):
        """ Returns dict of all packages matched. """
        if self._all_images:
            return self._all_images

        used_images = self.get_container_images()
        image_list = self.cli.docker_images()
        if not image_list:
            return None

        all_exprs = self.core_image_exprs + self.other_image_exprs
        for line in image_list:
            for image in all_exprs:
                fullname, shortname, version = self._match_image(image, line)
                if shortname is None:
                    continue

                if (fullname, version) not in used_images:
                    continue

                if image in self.core_image_exprs:
                    self._core_images[shortname] = version
                else:
                    self._other_images[shortname] = version

        # ensure sorted
        self._core_images = sorted_dict(self._core_images)
        self._other_images = sorted_dict(self._other_images)
        combined = {}
        combined.update(self._core_images)
        combined.update(self._other_images)
        self._all_images = sorted_dict(combined)

        return self._all_images

    @property
    def all(self):
        """
        Returns list of all images matched.
        """
        return self._all

    @property
    def core(self):
        """
        Only return results that matched from the "core" set of images.
        """
        if self._core_images:
            return self._core_images

        if self._other_images:
            return {}

        # go fetch
        _ = self.all
        return self._core_images


class APTPackageHelper(PackageHelperBase):
    """ Helpers for analysing apt packages. """
    def __init__(self, core_pkgs, other_pkgs=None):
        """
        @param core_pkgs: list of python.re expressions used to match
        package names.
        @param other_pkg_exprs: optional list of python.re expressions used to
        match packages that are not considered part of the core set. This can
        be used to distinguish between core and non-core packages.
        """
        self.core_pkg_exprs = core_pkgs
        self.other_pkg_exprs = other_pkgs or []
        self._core_packages = {}
        self._other_packages = {}
        self._all_packages = {}
        # Match any installed package status
        self._match_expr_template = r"^.i\s+({}[0-9a-z\-]*)\s+(\S+)\s+.+"
        self.cli = CLIHelper()

    def is_installed(self, pkg, allow_full_search=False):
        """
        Check if package is installed and return True/False. By default only
        checks against packages matched using main expression.

        @param pkg: name of package we want to check
        @param allow_full_search: if True will do a full dpkg search for the
                                  package of not found matches from the main
                                  expression.
        """
        if pkg in self.all:
            return True

        if not allow_full_search:
            return False

        dpkg_l = self.cli.dpkg_l()
        if not dpkg_l:
            return False

        cexpr = re.compile(rf"^.i\s+{pkg}\s+.+")
        for line in dpkg_l:
            # See https://man7.org/linux/man-pages/man1/dpkg-query.1.html for
            # package states.
            # Here we check package status not desired action.
            if cexpr.search(line):
                return True

        return False

    def get_version(self, pkg):
        """ Return version of package. """
        if pkg in self._all:
            return self._all[pkg]

        dpkg_l = self.cli.dpkg_l()
        if dpkg_l:
            for line in dpkg_l:
                name, version = self._match_package(pkg, line)
                if name:
                    return version

        return None

    def _match_package(self, pkg, entry):
        """ Returns tuple of (package name, version) """
        expr = self._match_expr_template.format(pkg)
        ret = re.compile(expr).match(entry)
        if ret:
            return ret[1], ret[2]

        return None, None

    @property
    def _all(self):
        """ Returns dict of all packages matched. """
        if self._all_packages:
            return self._all_packages

        dpkg_l = self.cli.dpkg_l()
        if not dpkg_l:
            return self._all_packages
        all_exprs = self.core_pkg_exprs + self.other_pkg_exprs
        for line in dpkg_l:
            for pkg in all_exprs:
                name, version = self._match_package(pkg, line)
                if name is None:
                    continue
                if pkg in self.core_pkg_exprs:
                    self._core_packages[name] = version
                else:
                    self._other_packages[name] = version

        # ensure sorted
        self._core_packages = sorted_dict(self._core_packages)
        self._other_packages = sorted_dict(self._other_packages)
        combined = {}
        combined.update(self._core_packages)
        combined.update(self._other_packages)
        self._all_packages = sorted_dict(combined)

        return self._all_packages

    @property
    def all(self):
        """
        Returns list of all packages matched. Each entry in the list is an item
        looking like "<pkgname> <pkgver>".
        """
        return self._all

    @property
    def core(self):
        """
        Only return results that matched from the "core" set of packages.
        """
        if self._core_packages:
            return self._core_packages

        # If _other_packages has contents it implies that we have already
        # collected and there are no core packages so return empty.
        if self._other_packages:
            return self._core_packages

        # go fetch
        _ = self.all
        return self._core_packages


@dataclass(frozen=True)
class AptPackage:
    """ Representation of an APT package.  """
    name: str
    version: str


class AptFactory(FactoryBase):
    """
    Factory to dynamically get package versions.

    AptPackage object is returned when a getattr() is done on this object
    using the name of package.
    """

    def __getattr__(self, name):
        log.debug("creating AptPackage object for %s", name)
        helper = APTPackageHelper([name])
        if name in helper.all:
            return AptPackage(name, helper.all[name])

        return None


class SnapPackageHelper(PackageHelperBase):
    """ Helpers for analysing snap packages. """
    def __init__(self, core_snaps, other_snaps=None):
        """
        @param core_snaps: list of python.re expressions used to match
        snap names.
        @param other_snap_exprs: optional list of python.re expressions used to
        match snap names for snaps that are not considered part of the
        core set.
        """
        self.core_snap_exprs = core_snaps
        self.other_snap_exprs = other_snaps or []
        self._core_snaps = {}
        self._other_snaps = {}
        self._all_snaps = {}
        self._match_expr_template = r"^({})\s+([\S]+)\s+([\d]+)\s+(\S+)\s+.+"
        self.snap_list_all = CLIHelper().snap_list_all()

    def is_installed(self, pkg):
        return pkg in self.all

    @staticmethod
    def _get_snap_info_from_line(line, cexpr):
        """ Returns snap name and info if found in line. """
        ret = re.match(cexpr, line)
        if ret:
            return {'name': ret.group(1),
                    'version': ret.group(2),
                    'revision': ret.group(3),
                    'channel': ret.group(4),
                    }

        return None

    def get_revision(self, snap):
        """ Return revision of package.
        """
        info = self._get_snap_info(snap)
        if info:
            return info[0]['revision']

        return None

    def get_version(self, pkg):
        """ Return version of snap package.

        Assumes only one snap will be matched.
        """
        info = self._get_snap_info(pkg)
        if info:
            return info[0]['version']

        return None

    def get_channel(self, pkg):
        """ Return channel of snap package.

        Assumes only one snap will be matched.
        """
        info = self._get_snap_info(pkg)
        if info:
            return info[0]['channel']

        return None

    def _get_snap_info(self, snap_name_expr):
        """
        Return a list of info for snaps matched using the expression.

        @param snap_name_expr: is a regular expression that can match one or
                               more snaps.
        @return: a list of snaps and their info.
        """
        if not self.snap_list_all:
            return None

        info = []
        cexpr = re.compile(self._match_expr_template.format(snap_name_expr))
        for line in self.snap_list_all:
            snap_info = self._get_snap_info_from_line(line, cexpr)
            if snap_info:
                info.append(snap_info)

        return info

    @property
    def _all(self):
        if self._all_snaps:
            return self._all_snaps

        if not self.snap_list_all:
            return {}

        _core = {}
        _other = {}
        all_exprs = self.core_snap_exprs + self.other_snap_exprs
        for snap in all_exprs:
            for snap_info in self._get_snap_info(snap):
                name = snap_info['name']
                version = snap_info['version']
                channel = snap_info['channel']
                # only show latest version installed
                if snap in self.core_snap_exprs:
                    if name in _core:
                        if version > _core[name]['version']:
                            _core[name]['version'] = version
                            _core[name]['channel'] = channel
                    else:
                        _core[name] = {'version': version,
                                       'channel': channel}
                else:
                    if name in _other:
                        if version > _other[name]['version']:
                            _other[name]['version'] = version
                            _other[name]['channel'] = channel
                    else:
                        _other[name] = {'version': version,
                                        'channel': channel}

        # ensure sorted
        self._core_snaps = sorted_dict(_core)
        self._other_snaps = sorted_dict(_other)
        combined = {}
        combined.update(_core)
        combined.update(_other)
        self._all_snaps = sorted_dict(combined)

        return self._all_snaps

    @property
    def all(self):
        """
        Return results that matched from the all/any set of snaps.
        """
        return self._all

    @property
    def all_formatted(self):
        _all = self.all
        if not _all:
            return []

        return [f"{name} {info['version']}" for name, info in _all.items()]

    @property
    def core(self):
        """
        Only return results that matched from the "core" set of snaps.
        """
        if self._core_snaps:
            return self._core_snaps

        if self._other_snaps:
            return {}

        # go fetch
        _ = self.all
        return self._core_snaps



import abc
import datetime
import json
import os
import pathlib
import re
import subprocess
import tempfile
from dataclasses import dataclass, field, fields
from functools import cached_property

from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers.common import (
    CmdOutput,
    get_ps_axo_flags_available,
    HostHelpersBase,
    reset_command,
    run_post_exec_hooks,
    run_pre_exec_hooks,
    SourceRunner,
)
from hotsos.core.host_helpers.exceptions import (
    catch_exceptions,
    CLI_COMMON_EXCEPTIONS,
    CLIExecError,
    CommandNotFound,
    SourceNotFound,
)
from hotsos.core.log import log


@dataclass
class CmdBase:
    """ Base class for all command source types.

    Provides a way to save original state and restore to that state.
    """
    def __post_init__(self):
        self.hooks = {}
        # make a copy of field original values.
        for f in fields(self):
            setattr(self, 'original_' + f.name, f.type(getattr(self, f.name)))

    def get_original_attr_value(self, name):
        return getattr(self, 'original_' + name)

    def reset(self):
        """ Reset fields to original values. """
        for f in fields(self):
            setattr(self, f.name, f.type(getattr(self, 'original_' + f.name)))

    @classmethod
    def safe_readlines(cls, path):
        with open(path, 'r', encoding='utf-8', errors="surrogateescape") as fd:
            return fd.readlines()

    def register_hook(self, name, f):
        """
        Implementations of this class can register hooks to
        be run when the __call__ method is executed. Currently
        supported hooks include:

            * pre-exec - run before __call__ method.
            * post-exec - run after __call__ method and take its output as
                          input.
        """
        self.hooks[name] = f


@dataclass
class FileCmdBase(CmdBase):
    """
    State used to execute a file-based command i.e. a command whose output is
    already saved in a file.
    """
    path: str
    json_decode: bool = False
    singleline: bool = False
    decode_error_handling: str = None

    def __post_init__(self):
        self.path = os.path.join(HotSOSConfig.data_root, self.path)
        super().__post_init__()


@dataclass
class BinCmdBase(CmdBase):
    """ State used to execute a binary command. """
    cmd: str
    json_decode: bool = False
    singleline: bool = False
    cmd_extras: list = field(default_factory=lambda: [])


class BinCmd(BinCmdBase):
    """ Implements binary command execution. """
    TYPE = "BIN"

    @catch_exceptions(*CLI_COMMON_EXCEPTIONS)
    @reset_command
    @run_post_exec_hooks
    @run_pre_exec_hooks
    def __call__(self, *args, skip_json_decode=False, **kwargs):
        cmd = self.cmd
        if args:
            cmd = cmd.format(*args)

        if kwargs:
            cmd = cmd.format(**kwargs)

        _cmd = cmd.split() + self.cmd_extras
        out = subprocess.run(_cmd, timeout=HotSOSConfig.command_timeout,
                             capture_output=True, check=False)
        output = out.stdout
        if out.stderr:
            output += out.stderr

        if out.returncode != 0:
            log.info("command '%s' exited with non-zero code '%s'", cmd,
                     out.returncode)

        try:
            output = output.decode('UTF-8')
        except UnicodeDecodeError:
            log.exception("failed to decode command output for '%s'", cmd)
            output = ''

        if self.json_decode and not skip_json_decode:
            return CmdOutput(json.loads(output))

        if self.singleline:
            return CmdOutput(output.strip())

        return CmdOutput(output.splitlines(keepends=True))


class FileCmd(FileCmdBase):
    """ Implements file-based command execution.

    This is used e.g. with sosreports where the output of a command is saved
    to disk.
    """
    TYPE = "FILE"

    @catch_exceptions(*CLI_COMMON_EXCEPTIONS)
    @reset_command
    @run_post_exec_hooks
    @run_pre_exec_hooks
    def __call__(self, *args, skip_load_contents=False, **kwargs):
        if args:
            self.path = self.path.format(*args)

        if kwargs:
            self.path = self.path.format(**kwargs)

        if not os.path.exists(self.path):
            raise SourceNotFound(self.path)

        if skip_load_contents:
            CmdOutput(None, self.path)

        # NOTE: any post-exec hooks must be aware that their input will be
        # defined by the following.
        if self.json_decode:
            with open(self.path, encoding='utf-8') as fd:
                output = json.load(fd)
        else:
            output = []
            ln = 0
            with open(self.path, 'rb') as fd:
                for line in fd:
                    ln += 1
                    try:
                        decode_kwargs = {}
                        if self.decode_error_handling:
                            decode_kwargs['errors'] = \
                                                     self.decode_error_handling

                        _out = line.decode(**decode_kwargs)
                        output.append(_out)
                    except UnicodeDecodeError:
                        # maintain line count but store empty line.
                        # we could in the future consider other decode options
                        # as a fallback.
                        output.append('')
                        log.exception("failed to decode line %s "
                                      "(decode_error_handling=%s)", ln,
                                      self.decode_error_handling)

                    if self.singleline:
                        break

            if self.singleline:
                return CmdOutput(output[0].strip(), self.path)

        return CmdOutput(output, self.path)


class BinFileCmd(FileCmd):
    """ This is used when we are executing an actual binary/command against a
    file. """

    @catch_exceptions(*CLI_COMMON_EXCEPTIONS)
    @reset_command
    @run_post_exec_hooks
    @run_pre_exec_hooks
    def __call__(self, *args, **kwargs):
        # NOTE: we check the original path since by this point the 'path'
        # attribute will have been modified to include a binary and other args
        # required to execute it.
        if not os.path.exists(self.get_original_attr_value('path')):
            raise SourceNotFound(self.get_original_attr_value('path'))

        if args:
            self.path = self.path.format(*args)

        if kwargs:
            self.path = self.path.format(**kwargs)

        # If this file is part of a sosreport we want to make sure it is run
        # in the same timezone context as the sosreport host.
        env = {}
        try:
            env['TZ'] = DateFileCmd('sos_commands/date/date',
                                    singleline=True)(format="+%Z").value
        except SourceNotFound:
            pass

        # Now split into a command and run
        output = subprocess.check_output(self.path.split(),
                                         timeout=HotSOSConfig.command_timeout,
                                         stderr=subprocess.STDOUT, env=env)

        output = output.decode('UTF-8')
        return CmdOutput(output.splitlines(keepends=True))


class JournalctlBase():
    """ Base class for journalctl command implementations. """
    @property
    def since_date(self):
        """
        Returns a string datetime to be used with journalctl --since. This time
        reflects the maximum depth of history we will search in the journal.

        The datetime value returned takes into account config from HotSOSConfig
        and has the format "YEAR-MONTH-DAY". It does not specify a time.
        """
        current = CLIHelper().date(format="--iso-8601")
        ts = datetime.datetime.strptime(current, "%Y-%m-%d")
        if HotSOSConfig.use_all_logs:
            days = HotSOSConfig.max_logrotate_depth
        else:
            days = 1

        ts = ts - datetime.timedelta(days=days)
        return ts.strftime("%Y-%m-%d")


class JournalctlBinCmd(BinCmd, JournalctlBase):
    """ Implements binary journalctl command. """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.register_hook("pre-exec", self.format_journalctl_cmd)

    def format_journalctl_cmd(self, **kwargs):
        """ Add optional extras to journalctl command. """
        if kwargs.get("unit"):
            self.cmd = f"{self.cmd} --unit {kwargs.get('unit')}"

        if kwargs.get("date"):
            self.cmd = f"{self.cmd} --since {kwargs.get('date')}"
        else:
            self.cmd = f"{self.cmd} --since {self.since_date}"


class JournalctlBinFileCmd(BinFileCmd, JournalctlBase):
    """ Implements file-based journalctl command. """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.register_hook("pre-exec", self.preformat_sos_journalctl)

    def preformat_sos_journalctl(self, **kwargs):
        self.path = f"journalctl -oshort-iso -D {self.path}"
        if kwargs.get("unit"):
            self.path = f"{self.path} --unit {kwargs.get('unit')}"

        if kwargs.get("date"):
            self.path = f"{self.path} --since {kwargs.get('date')}"
        else:
            self.path = f"{self.path} --since {self.since_date}"


class OVSAppCtlBinCmd(BinCmd):
    """ Implements ovs-appctl binary command. """
    def __call__(self, *args, **kwargs):
        # Set defaults for optional args
        for key in ['flags', 'args']:
            if key not in kwargs:
                kwargs[key] = ''

        return super().__call__(*args, **kwargs)


class OVSAppCtlFileCmd(FileCmd):
    """ Implements ovs-appctl file-based command. """
    def __call__(self, *args, **kwargs):
        for key in ['flags', 'args']:
            if key in kwargs:
                kwargs[key] = f'_{kwargs[key]}'
            else:
                kwargs[key] = ''

        if 'args' in kwargs and kwargs['command'].startswith('dpctl/'):
            # e.g. this would be for dpctl datapath
            kwargs['args'] = kwargs['args'].replace('@', '_')

        kwargs['command'] = kwargs['command'].replace('/', '.')
        try:
            out = super().__call__(*args, **kwargs)
        except SourceNotFound:
            log.debug("%s: source not found: %s", self.__class__.__name__,
                      self.path)
            raise

        return out


OFPROTOCOL_VERSIONS = [
    "OpenFlow15",
    "OpenFlow14",
    "OpenFlow13",
    "OpenFlow12",
    "OpenFlow11",
    "OpenFlow10",
]


class OVSOFCtlBinCmd(BinCmd):
    """ Implementation of ovs-ofctl binary command. """
    def __call__(self, *args, **kwargs):
        """
        First try without specifying protocol version. If error is raised
        try with different versions until we get a result.
        """
        self.cmd = f"ovs-ofctl {self.cmd}"
        try:
            return super().__call__(*args, **kwargs)
        except CLIExecError:
            log.debug("%s: command with no protocol version failed",
                      self.__class__.__name__)

        # If the command raised an exception it will have been caught by the
        # catch_exceptions decorator and [] returned. We have no way of knowing
        # if that was the actual return or an exception was raised so we just
        # go ahead and retry with specific OF versions until we get a result.
        for ver in OFPROTOCOL_VERSIONS:
            log.debug("%s: trying again with protocol version %s",
                      self.__class__.__name__, ver)
            self.reset()
            self.cmd = f"ovs-ofctl -O {ver} {self.cmd}"
            try:
                return super().__call__(*args, **kwargs)
            except CLIExecError:
                log.debug("%s: command with protocol version %s failed",
                          self.__class__.__name__, ver)

        return CmdOutput([])


class OVSOFCtlFileCmd(FileCmd):
    """ Implementation of ovs-ofctl file-based command. """
    def __call__(self, *args, **kwargs):
        """
        We do this in reverse order to bin command since it won't actually
        raise an error.
        """
        for ver in OFPROTOCOL_VERSIONS:
            log.debug("%s: trying again with protocol version %s",
                      self.__class__.__name__, ver)
            self.reset()
            try:
                kwargs['ofversion'] = f'_-O_{ver}'
                out = super().__call__(*args, **kwargs)
                if out.value and 'version negotiation failed' in out.value[0]:
                    continue

                return out
            except SourceNotFound:
                pass

        try:
            kwargs['ofversion'] = ''
            return super().__call__(*args, **kwargs)
        except SourceNotFound:
            log.debug("%s: command with no protocol version failed",
                      self.__class__.__name__)

        return None


class DateBinCmd(BinCmd):
    """ Implementation binary date command. """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.register_hook("pre-exec", self.format_date_cmd)

    def format_date_cmd(self, **kwargs):
        """ Add formatting to date command. """
        no_format = kwargs.get('no_format', False)
        fmt = kwargs.get('format')
        if not no_format and fmt is None:
            fmt = '+%s'

        self.cmd = f'{self.cmd} --utc'
        if fmt:
            # this can't get split() so add to the end of the command list
            self.cmd_extras = [fmt]


class DateFileCmd(FileCmd):
    """ Implementation of file-based date command. """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.register_hook("post-exec", self.format_date)

    def format_date(self, output, **kwargs):
        """ Apply some post-processing to the date output.

        @param output: CmdOutput object
        """
        no_format = kwargs.get('no_format', False)
        fmt = kwargs.get('format')
        if not no_format and fmt is None:
            fmt = '+%s'

        ret = re.match(r"^(\S+ \S*\s*[0-9]+ [0-9:]+)\s*"
                       r"([A-Z]*|[+-]?[0-9]*)?"
                       r"\s*([0-9]+)$",
                       output.value)

        if ret is None:
            log.error("%s has invalid date string '%s'", self.path,
                      output.value)
            return CmdOutput('', self.path)

        tz = ret[2]
        # NOTE: date command doesn't recognise HKT for some reason so we
        # convert to a format that is recognised.
        if tz == 'HKT':
            tz = 'UTC+8'

        # Include tz if available and then convert to utc
        date = f"{ret[1]} {tz} {ret[3]}"
        cmd = ["date", "--utc", f"--date={date}"]
        if fmt:
            cmd.append(fmt)

        output = subprocess.check_output(cmd,
                                         timeout=HotSOSConfig.command_timeout)
        # date sometimes adds multiple whitespaces between fields so collapse
        # them.
        output = re.compile(r"\s+").sub(' ', output.decode('UTF-8'))
        ret = output.splitlines(keepends=True)[0]
        # always singleline so always strip trailing newline
        return CmdOutput(ret.strip(), self.path)


class CephJSONFileCmd(FileCmd):
    """
    Some ceph commands that use --format json have some extra text added to the
    end of the file (typically from stderr) which causes it to be invalid json
    so we have to strip that final line before decoding the contents.
    """
    def __init__(self, *args, first_line_filter=None, last_line_filter=None,
                 **kwargs):
        super().__init__(*args, **kwargs)
        if first_line_filter or last_line_filter:
            self.register_hook('pre-exec', self.format_json_contents)
            self.register_hook('post-exec', self.cleanup)
            self.orig_path = None
            self.first_line_filter = first_line_filter
            self.last_line_filter = last_line_filter

    def format_json_contents(self, *_args, **_kwargs):
        if not os.path.exists(self.path):
            raise SourceNotFound(self.path)

        with open(self.path, encoding='utf-8') as f:
            lines = f.readlines()

        if self.first_line_filter:
            line_filter = self.first_line_filter
        else:
            line_filter = self.last_line_filter

        if lines and lines[-1].startswith(line_filter):
            lines = lines[:-1]
            with tempfile.NamedTemporaryFile(mode='w+t', delete=False) as tmp:
                tmp.write(''.join(lines))
                tmp.close()
                self.orig_path = self.path
                self.path = tmp.name

    def cleanup(self, output, **_kwargs):
        """
        @param output: CmdOutput object
        """
        if self.orig_path:
            os.unlink(self.path)
            self.path = self.orig_path
            self.orig_path = None

        return output


class CLICacheWrapper():
    """ Wrapper for cli cache. """
    def __init__(self, cache_load_f, cache_save_f):
        self.load_f = cache_load_f
        self.save_f = cache_save_f

    def load(self, key):
        return self.load_f(key)

    def save(self, key, value):
        return self.save_f(key, value)


class CLIHelperBase(HostHelpersBase):
    """ Base class for clihelper implementations. """
    def __init__(self):
        self._command_catalog = None
        super().__init__()
        self.cli_cache = CLICacheWrapper(self.cache_load, self.cache_save)

    @property
    def cache_root(self):
        """ Cache at plugin level rather than globally. """
        return HotSOSConfig.plugin_tmp_dir

    @property
    def cache_type(self):
        return 'cli'

    @property
    def cache_name(self):
        return "commands"

    def cache_load(self, key):
        return self.cache.get(key)

    def cache_save(self, key, value):
        return self.cache.set(key, value)

    @property
    def command_catalog(self):
        if self._command_catalog:
            return self._command_catalog

        self._command_catalog = {
            'apt_config_dump':
                [BinCmd('apt-config dump'),
                 FileCmd('sos_commands/apt/apt-config_dump')],
            'apparmor_status':
                [BinCmd('apparmor_status'),
                 FileCmd('sos_commands/apparmor/apparmor_status')],
            'ceph_daemon_osd_config_show':
                [BinCmd('ceph daemon osd.{osd_id} config show',
                        json_decode=True),
                 # requires sosreport 4.3 or above
                 FileCmd('sos_commands/ceph_osd/'
                         'ceph_daemon_osd.{osd_id}_config_show',
                         json_decode=True)],
            'ceph_daemon_osd_dump_mempools':
                [BinCmd('ceph daemon osd.{osd_id} dump mempools',
                        json_decode=True),
                 # requires sosreport 4.3 or above
                 FileCmd('sos_commands/ceph_osd/'
                         'ceph_daemon_osd.{osd_id}_dump_mempools',
                         json_decode=True)],
            'ceph_health_detail_json_decoded':
                [BinCmd('ceph health detail --format json-pretty',
                        json_decode=True),
                 # sosreport < 4.2
                 FileCmd('sos_commands/ceph/json_output/'
                         'ceph_health_detail_--format_json-pretty',
                         json_decode=True),
                 # sosreport >= 4.2
                 FileCmd('sos_commands/ceph_mon/json_output/'
                         'ceph_health_detail_--format_json-pretty',
                         json_decode=True)],
            'ceph_mon_dump_json_decoded':
                [BinCmd('ceph mon dump --format json-pretty',
                        json_decode=True),
                 # sosreport < 4.2
                 CephJSONFileCmd('sos_commands/ceph/json_output/'
                                 'ceph_mon_dump_--format_json-pretty',
                                 json_decode=True,
                                 first_line_filter='dumped monmap epoch',
                                 last_line_filter='dumped monmap epoch'),
                 # sosreport >= 4.2
                 CephJSONFileCmd('sos_commands/ceph_mon/json_output/'
                                 'ceph_mon_dump_--format_json-pretty',
                                 json_decode=True,
                                 first_line_filter='dumped monmap epoch',
                                 last_line_filter='dumped monmap epoch')],
            'ceph_osd_dump_json_decoded':
                [BinCmd('ceph osd dump --format json-pretty',
                        json_decode=True),
                 # sosreport < 4.2
                 FileCmd('sos_commands/ceph/json_output/'
                         'ceph_osd_dump_--format_json-pretty',
                         json_decode=True),
                 # sosreport >= 4.2
                 FileCmd('sos_commands/ceph_mon/json_output/'
                         'ceph_osd_dump_--format_json-pretty',
                         json_decode=True)],
            'ceph_df_json_decoded':
                [BinCmd('ceph df --format json-pretty',
                        json_decode=True),
                 # sosreport < 4.2
                 FileCmd('sos_commands/ceph/json_output/'
                         'ceph_df_--format_json-pretty',
                         json_decode=True),
                 # sosreport >= 4.2
                 FileCmd('sos_commands/ceph_mon/json_output/'
                         'ceph_df_--format_json-pretty',
                         json_decode=True)],
            'ceph_osd_df_tree_json_decoded':
                [BinCmd('ceph osd df tree --format json-pretty',
                        json_decode=True),
                 # sosreport < 4.2
                 FileCmd('sos_commands/ceph/json_output/'
                         'ceph_osd_df_tree_--format_json-pretty',
                         json_decode=True),
                 # sosreport >= 4.2
                 FileCmd('sos_commands/ceph_mon/json_output/'
                         'ceph_osd_df_tree_--format_json-pretty',
                         json_decode=True)],
            'ceph_osd_crush_dump_json_decoded':
                [BinCmd('ceph osd crush dump', json_decode=True),
                 # sosreport < 4.2
                 FileCmd('sos_commands/ceph/ceph_osd_crush_dump',
                         json_decode=True),
                 # sosreport >= 4.2
                 FileCmd('sos_commands/ceph_mon/ceph_osd_crush_dump',
                         json_decode=True),
                 ],
            'ceph_pg_dump_json_decoded':
                [BinCmd('ceph pg dump --format json-pretty',
                        json_decode=True),
                 # sosreport < 4.2
                 CephJSONFileCmd('sos_commands/ceph/json_output/'
                                 'ceph_pg_dump_--format_json-pretty',
                                 json_decode=True,
                                 first_line_filter='dumped all',
                                 last_line_filter='dumped all'),
                 # sosreport >= 4.2
                 CephJSONFileCmd('sos_commands/ceph_mon/json_output/'
                                 'ceph_pg_dump_--format_json-pretty',
                                 json_decode=True,
                                 first_line_filter='dumped all',
                                 last_line_filter='dumped all')],
            'ceph_status_json_decoded':
                [BinCmd('ceph status --format json-pretty', json_decode=True),
                 # sosreport < 4.2
                 FileCmd('sos_commands/ceph/json_output/'
                         'ceph_status_--format_json-pretty',
                         json_decode=True),
                 # sosreport >= 4.2
                 FileCmd('sos_commands/ceph_mon/'
                         'json_output/ceph_status_--format_json-pretty',
                         json_decode=True),
                 ],
            'ceph_versions':
                [BinCmd('ceph versions'),
                 # sosreport < 4.2
                 FileCmd('sos_commands/ceph/ceph_versions'),
                 # sosreport >= 4.2
                 FileCmd('sos_commands/ceph_mon/ceph_versions'),
                 ],
            'ceph_volume_lvm_list':
                [BinCmd('ceph-volume lvm list'),
                 # sosreport < 4.2
                 FileCmd('sos_commands/ceph/ceph-volume_lvm_list'),
                 # sosreport >= 4.2
                 FileCmd('sos_commands/ceph_osd/ceph-volume_lvm_list'),
                 ],
            'ceph_report_json_decoded':
                [BinCmd('ceph report', json_decode=True),
                 # sosreport < 4.2
                 CephJSONFileCmd('sos_commands/ceph/ceph_report',
                                 json_decode=True,
                                 first_line_filter='report',
                                 last_line_filter='report'),
                 # sosreport >= 4.2
                 CephJSONFileCmd('sos_commands/ceph_mon/ceph_report',
                                 json_decode=True,
                                 first_line_filter='report',
                                 last_line_filter='report'),
                 ],
            'ceph_mgr_module_ls':
                [BinCmd('ceph mgr module ls', json_decode=True),
                 # sosreport < 4.2
                 CephJSONFileCmd('sos_commands/ceph/ceph_mgr_module_ls',
                                 json_decode=True),
                 # sosreport >= 4.2
                 CephJSONFileCmd('sos_commands/ceph_mon/ceph_mgr_module_ls',
                                 json_decode=True),
                 ],
            'date':
                [DateBinCmd('date', singleline=True),
                 DateFileCmd('sos_commands/date/date', singleline=True),
                 # this is for legacy sosreport versions
                 DateFileCmd('sos_commands/general/date', singleline=True)],
            'df':
                [BinCmd('df'),
                 FileCmd('df')],
            'dmesg':
                [BinCmd('dmesg'),
                 FileCmd('sos_commands/kernel/dmesg')],
            'docker_images':
                [BinCmd('docker images'),
                 FileCmd('sos_commands/docker/docker_images')],
            'docker_ps':
                [BinCmd('docker ps'),
                 FileCmd('sos_commands/docker/docker_ps')],
            'dpkg_l':
                [BinCmd('dpkg -l'),
                 FileCmd('sos_commands/dpkg/dpkg_-l',
                         decode_error_handling='surrogateecape')],
            'ethtool':
                [BinCmd('ethtool {interface}'),
                 FileCmd('sos_commands/networking/ethtool_{interface}')],
            'hostname':
                [BinCmd('hostname', singleline=True),
                 FileCmd('hostname', singleline=True)],
            'hostnamectl':
                [BinCmd('hostnamectl'),
                 FileCmd('sos_commands/host/hostnamectl_status')],
            'ip_netns':
                [BinCmd('ip netns'),
                 FileCmd('sos_commands/networking/ip_netns')],
            'ip_addr':
                [BinCmd('ip -d address'),
                 FileCmd('sos_commands/networking/ip_-d_address')],
            'ip_link':
                [BinCmd('ip -s -d link'),
                 FileCmd('sos_commands/networking/ip_-s_-d_link')],
            'journalctl':
                [JournalctlBinCmd('journalctl -oshort-iso'),
                 JournalctlBinFileCmd('var/log/journal')],
            'ls_lanR_sys_block':
                [BinCmd('ls -lanR /sys/block/'),
                 FileCmd('sos_commands/block/ls_-lanR_.sys.block')],
            'lscpu':
                [BinCmd('lscpu'),
                 FileCmd('sos_commands/processor/lscpu')],
            'lsof_Mnlc':
                [BinCmd('lsof +M -n -l -c ""'),
                 FileCmd('sos_commands/process/lsof_M_-n_-l_-c')],
            'lxd_buginfo':
                [BinCmd('lxd.buginfo'),
                 FileCmd('sos_commands/lxd/lxd.buginfo')],
            'numactl':
                [BinCmd('numactl --hardware'),
                 FileCmd('sos_commands/numa/numactl_--hardware')],
            'ns_ip_addr':
                [BinCmd('ip netns exec {namespace} ip address show'),
                 FileCmd('sos_commands/networking/'
                         'ip_netns_exec_{namespace}_ip_address_show'),
                 FileCmd('sos_commands/networking/namespaces/{namespace}/'
                         'ip_netns_exec_{namespace}_ip_-d_address_show')],
            'ovn_nbctl_show':
                [BinCmd('ovn-nbctl --no-leader-only show'),
                 FileCmd('sos_commands/ovn_central/'
                         'ovn-nbctl_--no-leader-only_show'),
                 # sosreport < 4.5
                 FileCmd('sos_commands/ovn_central/ovn-nbctl_show')],
            'ovn_sbctl_show':
                [BinCmd('ovn-sbctl --no-leader-only show'),
                 FileCmd('sos_commands/ovn_central/'
                         'ovn-sbctl_--no-leader-only_show'),
                 # sosreport < 4.5
                 FileCmd('sos_commands/ovn_central/ovn-sbctl_show')],
            'ovs_vsctl_get':
                [BinCmd('ovs-vsctl get {table} {record} {column}',
                        singleline=True),
                 FileCmd('sos_commands/openvswitch/ovs-vsctl_-t_5_get_'
                         '{table}_{record}_{column}', singleline=True)],
            'ovs_vsctl_list':
                [BinCmd('ovs-vsctl list {table}'),
                 FileCmd('sos_commands/openvswitch/'
                         'ovs-vsctl_-t_5_list_{table}')],
            'ovs_vsctl_list_br':
                [BinCmd('ovs-vsctl list-br'),
                 FileCmd('sos_commands/openvswitch/ovs-vsctl_-t_5_list-br')],
            'ovs_appctl':
                [OVSAppCtlBinCmd('ovs-appctl {command} {flags} {args}'),
                 OVSAppCtlFileCmd('sos_commands/openvswitch/ovs-appctl_'
                                  '{command}{flags}{args}')],
            'ovs_ofctl':
                [OVSOFCtlBinCmd('{command} {args}'),
                 OVSOFCtlFileCmd('sos_commands/openvswitch/'
                                 'ovs-ofctl{ofversion}_{command}_{args}')],
            'pacemaker_crm_status':
                [BinCmd('crm status'),
                 FileCmd('sos_commands/pacemaker/crm_status')],
            'pebble_services':
                [BinCmd('pebble services'),
                 # This is how operator charms run it
                 BinCmd('/charm/bin/pebble services'),
                 # The following does not exist in sosreport yet but adding
                 # since it is useful for testing and will hopefully be
                 # supported in sos at some point.
                 FileCmd('sos_commands/pebble/pebble_services')],
            'ps':
                [BinCmd('ps auxwww'),
                 FileCmd('ps')],
            'ps_axo_flags':
                [BinCmd('ps axo flags,state,uid,pid,ppid,pgid,sid,cls,'
                        'pri,addr,sz,wchan:20,lstart,tty,time,cmd'),
                 # Older sosrepot uses 'wchan' option while newer ones use
                 # 'wchan:20' - thus the glob is to cover both
                 FileCmd(get_ps_axo_flags_available() or "")],
            'pro_status':
                [BinCmd('ua status'),
                 FileCmd('sos_commands/ubuntu/ua_status')],
            'rabbitmqctl_report':
                [BinCmd('rabbitmqctl report'),
                 FileCmd('sos_commands/rabbitmq/rabbitmqctl_report')],
            'snap_list_all':
                [BinCmd('snap list --all'),
                 FileCmd('sos_commands/snap/snap_list_--all'),
                 # sos legacy
                 FileCmd('sos_commands/snappy/snap_list_--all')],
            'sysctl_all':
                [BinCmd('sysctl -a'),
                 FileCmd('sos_commands/kernel/sysctl_-a')],
            'systemctl_status_all':
                [BinCmd('systemctl status --all'),
                 FileCmd('sos_commands/systemd/systemctl_status_--all',
                         decode_error_handling='backslashreplace')],
            'systemctl_list_units':
                [BinCmd('systemctl list-units'),
                 FileCmd('sos_commands/systemd/systemctl_list-units')],
            'systemctl_list_unit_files':
                [BinCmd('systemctl list-unit-files'),
                 FileCmd('sos_commands/systemd/systemctl_list-unit-files')],
            'udevadm_info_dev':
                [BinCmd('udevadm info /dev/{device}'),
                 FileCmd('sos_commands/block/udevadm_info_.dev.{device}')],
            'udevadm_info_exportdb':
                [BinCmd('udevadm info --export-db'),
                 FileCmd('sos_commands/devices/udevadm_info_--export-db')],
            'uname':
                [BinCmd('uname -a', singleline=True),
                 FileCmd('sos_commands/kernel/uname_-a', singleline=True)],
            'uptime':
                [BinCmd('uptime', singleline=True),
                 FileCmd('uptime', singleline=True)],
        }
        return self._command_catalog

    @abc.abstractmethod
    def __getattr__(self, cmdname):
        """ This is how commands are run. The command is looked up in the
        catalog and it's runner object is returned. The caller is expetced to
        call() the returned object to execute the command.

        @param cmdname: name of command we want to execute. This must match a
        name used to register a handler in the catalog.
        @return: SourceRunner object.
        """


class CLIHelper(CLIHelperBase):
    """
    This is used when we want to have command output as the return value when
    a command is executed.
    """

    def __getattr__(self, cmdname):
        try:
            return SourceRunner(cmdname, self.command_catalog[cmdname],
                                self.cli_cache)
        except KeyError as exc:
            raise CommandNotFound(cmdname, exc) from exc


class CLIHelperFile(CLIHelperBase):
    """
    This is used when we want the return value of a command to be a path to a
    file containing the return value of executing that command.

    This will do one of two things; if the command output originates from a
    file e.g. a sosreport command output file, it will return the path to that
    file. If the command is executed as a binary, its output is written to a
    temporary file and the path to that file is returned.
    """

    def __init__(self, *args, delete_temp=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.delete_temp = delete_temp
        self._tmp_file_mtime = None

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        do_delete = (self.delete_temp or
                     self._tmp_file_mtime ==
                     os.path.getmtime(self.output_file))
        if do_delete:
            os.remove(self.output_file)

        # We want exceptions to be raised
        return False

    @cached_property
    def output_file(self):
        path = tempfile.mktemp(dir=HotSOSConfig.plugin_tmp_dir)
        pathlib.Path(path).touch()
        self._tmp_file_mtime = os.path.getmtime(path)
        return path

    def __getattr__(self, cmdname):
        try:
            ret = SourceRunner(cmdname, self.command_catalog[cmdname],
                               self.cli_cache, output_file=self.output_file)
            return ret
        except KeyError as exc:
            raise CommandNotFound(cmdname, exc) from exc

        return None



import os
import re
from functools import cached_property

from hotsos.core import host_helpers, plugintools
from hotsos.core.config import HotSOSConfig
from hotsos.core.plugins.kernel.config import KernelConfig


class KernelBase():
    """ Base class for kernel plugin helpers. """
    @cached_property
    def version(self):
        """Returns string kernel version."""
        uname = host_helpers.CLIHelper().uname()
        if uname:
            ret = re.compile(r"^Linux\s+\S+\s+(\S+)\s+.+").match(uname)
            if ret:
                return ret[1]

        return ""

    @cached_property
    def isolcpus_enabled(self):
        return KernelConfig().get('isolcpus') is not None

    @cached_property
    def boot_parameters(self):
        """Returns list of boot parameters."""
        parameters = []
        path = os.path.join(HotSOSConfig.data_root, "proc/cmdline")
        if os.path.exists(path):
            with open(path, encoding='utf-8') as fd:
                cmdline = fd.read().strip()
            for entry in cmdline.split():
                if entry.startswith("BOOT_IMAGE"):
                    continue

                if entry.startswith("root="):
                    continue

                parameters.append(entry)

            return parameters

        return None


class KernelChecks(KernelBase, plugintools.PluginPartBase):
    """ Base class for all kernel checks. """
    plugin_name = 'kernel'
    plugin_root_index = 15

    @classmethod
    def is_runnable(cls):
        """
        Determine whether or not this plugin can and should be run.

        @return: True or False
        """
        return True



# flake8: noqa

# From https://github.com/openstack/manila/blob/master/manila/exception.py
# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' manila/exception.py
MANILA_EXCEPTIONS = [
    "ConvertedException",
    "Error",
    "ManilaException",
    "NetworkException",
    "NetworkBindException",
    "NetworkBadConfigurationException",
    "BadConfigurationException",
    "NotAuthorized",
    "AdminRequired",
    "PolicyNotAuthorized",
    "Conflict",
    "Invalid",
    "InvalidRequest",
    "InvalidResults",
    "InvalidInput",
    "InvalidContentType",
    "InvalidHost",
    "InvalidParameterValue",
    "InvalidUUID",
    "InvalidDriverMode",
    "InvalidAPIVersionString",
    "VersionNotFoundForAPIMethod",
    "InvalidGlobalAPIVersion",
    "InvalidCapacity",
    "NotFound",
    "MessageNotFound",
    "Found",
    "InUse",
    "AvailabilityZoneNotFound",
    "ShareNetworkNotFound",
    "ShareNetworkSubnetNotFound",
    "ShareServerNotFound",
    "ShareServerNotFoundByFilters",
    "InvalidShareNetwork",
    "ShareServerInUse",
    "ShareServerMigrationError",
    "ShareServerMigrationFailed",
    "InvalidShareServer",
    "ShareMigrationError",
    "ShareMigrationFailed",
    "ShareDataCopyFailed",
    "ShareDataCopyCancelled",
    "ServiceIPNotFound",
    "AdminIPNotFound",
    "ShareServerNotCreated",
    "ShareServerNotReady",
    "ServiceNotFound",
    "ServiceIsDown",
    "HostNotFound",
    "SchedulerHostFilterNotFound",
    "SchedulerHostWeigherNotFound",
    "HostBinaryNotFound",
    "InvalidReservationExpiration",
    "InvalidQuotaValue",
    "QuotaNotFound",
    "QuotaExists",
    "QuotaResourceUnknown",
    "ProjectUserQuotaNotFound",
    "ProjectShareTypeQuotaNotFound",
    "ProjectQuotaNotFound",
    "QuotaClassNotFound",
    "QuotaUsageNotFound",
    "ReservationNotFound",
    "OverQuota",
    "MigrationNotFound",
    "MigrationNotFoundByStatus",
    "FileNotFound",
    "MigrationError",
    "MalformedRequestBody",
    "ConfigNotFound",
    "PasteAppNotFound",
    "NoValidHost",
    "WillNotSchedule",
    "QuotaError",
    "ShareSizeExceedsAvailableQuota",
    "SnapshotSizeExceedsAvailableQuota",
    "ShareSizeExceedsLimit",
    "ShareLimitExceeded",
    "SnapshotLimitExceeded",
    "ShareNetworksLimitExceeded",
    "ShareGroupsLimitExceeded",
    "ShareGroupSnapshotsLimitExceeded",
    "ShareReplicasLimitExceeded",
    "ShareReplicaSizeExceedsAvailableQuota",
#    "GlusterfsException",
    "InvalidShare",
    "ShareBusyException",
    "InvalidShareInstance",
    "ManageInvalidShare",
    "ManageShareServerError",
    "UnmanageInvalidShare",
    "PortLimitExceeded",
    "ShareAccessExists",
    "ShareAccessMetadataNotFound",
    "ShareSnapshotAccessExists",
    "InvalidSnapshotAccess",
    "InvalidShareAccess",
    "InvalidShareAccessLevel",
    "InvalidShareAccessType",
    "ShareBackendException",
    "ExportLocationNotFound",
    "ShareNotFound",
    "ShareSnapshotNotFound",
    "ShareSnapshotInstanceNotFound",
    "ShareSnapshotNotSupported",
    "ShareGroupSnapshotNotSupported",
    "ShareSnapshotIsBusy",
    "InvalidShareSnapshot",
    "InvalidShareSnapshotInstance",
    "ManageInvalidShareSnapshot",
    "UnmanageInvalidShareSnapshot",
    "ShareMetadataNotFound",
    "InvalidMetadata",
    "InvalidMetadataSize",
    "SecurityServiceNotFound",
    "InvalidSecurityService",
    "ShareNetworkSecurityServiceAssociationError",
    "ShareNetworkSecurityServiceDissociationError",
    "SecurityServiceFailedAuth",
    "InvalidVolume",
    "InvalidShareType",
    "InvalidShareGroupType",
    "InvalidExtraSpec",
    "VolumeNotFound",
    "VolumeSnapshotNotFound",
    "ShareTypeNotFound",
    "ShareGroupTypeNotFound",
    "ShareTypeAccessNotFound",
    "ShareGroupTypeAccessNotFound",
    "ShareTypeNotFoundByName",
    "ShareGroupTypeNotFoundByName",
    "ShareTypeExtraSpecsNotFound",
    "ShareGroupTypeSpecsNotFound",
    "ShareTypeInUse",
    "IPAddressInUse",
    "ShareGroupTypeInUse",
    "ShareTypeExists",
    "ShareTypeDoesNotExist",
    "DefaultShareTypeNotConfigured",
    "ShareGroupTypeExists",
    "ShareTypeAccessExists",
    "ShareGroupTypeAccessExists",
    "ShareTypeCreateFailed",
    "ShareTypeUpdateFailed",
    "ShareGroupTypeCreateFailed",
    "ManageExistingShareTypeMismatch",
    "ShareExtendingError",
    "ShareShrinkingError",
    "ShareShrinkingPossibleDataLoss",
    "InstanceNotFound",
    "BridgeDoesNotExist",
    "ServiceInstanceException",
    "ServiceInstanceUnavailable",
    "StorageResourceException",
    "StorageResourceNotFound",
    "SnapshotResourceNotFound",
    "SnapshotUnavailable",
#    "NetAppException",
#    "VserverNotFound",
#    "VserverNotSpecified",
#    "VserverNotReady",
#    "EMCPowerMaxXMLAPIError",
#    "EMCPowerMaxLockRequiredException",
#    "EMCPowerMaxInvalidMoverID",
#    "EMCVnxXMLAPIError",
#    "EMCVnxLockRequiredException",
#    "EMCVnxInvalidMoverID",
#    "EMCUnityError",
#    "HPE3ParInvalidClient",
#    "HPE3ParInvalid",
#    "HPE3ParUnexpectedError",
#    "GPFSException",
#    "GPFSGaneshaException",
    "GaneshaCommandFailure",
    "InvalidSqliteDB",
    "SSHException",
#    "HDFSException",
#    "MapRFSException",
#    "ZFSonLinuxException",
    "QBException",
    "QBRpcException",
    "SSHInjectionThreat",
#    "HNASBackendException",
#    "HNASConnException",
#    "HNASSSCIsBusy",
#    "HNASSSCContextChange",
#    "HNASDirectoryNotEmpty",
#    "HNASItemNotFoundException",
#    "HNASNothingToCloneException",
    "ShareGroupNotFound",
    "ShareGroupSnapshotNotFound",
    "ShareGroupSnapshotMemberNotFound",
    "InvalidShareGroup",
    "InvalidShareGroupSnapshot",
    "DriverNotInitialized",
    "ShareResourceNotFound",
    "ShareUmountException",
    "ShareMountException",
    "ShareCopyDataException",
    "ReplicationException",
    "ShareReplicaNotFound",
#    "TegileAPIException",
    "StorageCommunicationException",
    "EvaluatorParseException",
#    "HSPBackendException",
#    "HSPTimeoutException",
#    "HSPItemNotFoundException",
#    "NexentaException",
    "LockCreationFailed",
    "LockingFailed",
    "GaneshaException",
#    "InfortrendCLIException",
#    "InfortrendNASException",
#    "ZadaraUnknownCmd",
#    "ZadaraSessionRequestException",
#    "ZadaraBadHTTPResponseStatus",
#    "ZadaraFailedCmdWithDump",
#    "ZadaraVPSANoActiveController",
#    "ZadaraServerCreateFailure",
#    "ZadaraAttachmentsNotFound",
#    "ZadaraManilaInvalidAccessKey",
#    "ZadaraVPSAVolumeShareFailed",
#    "ZadaraInvalidShareAccessType",
#    "ZadaraShareNotFound",
#    "ZadaraExtendShareFailed",
#    "ZadaraInvalidProtocol",
#    "ZadaraShareNotValid",
#    "ZadaraVPSASnapshotCreateFailed",
#    "ZadaraVPSASnapshotManageFailed",
#    "ZadaraServerNotFound",
]
